import keras as ks
from keras import ops


def _compute_fans(shape):
    """Computes the number of input and output units for a weight shape.

    Taken from original TensorFlow implementation and copied here for static reference.

    Args:
        shape: Integer shape tuple or tensor shape.

    Returns:
        A tuple of integer scalars (fan_in, fan_out).
    """
    if len(shape) < 1:  # Just to avoid errors for constants.
        fan_in = fan_out = 1
    elif len(shape) == 1:
        fan_in = fan_out = shape[0]
    elif len(shape) == 2:
        fan_in = shape[0]
        fan_out = shape[1]
    else:
        # Assuming convolution kernels (2D, 3D, or more).
        # kernel shape: (..., input_depth, depth)
        receptive_field_size = 1
        for dim in shape[:-2]:
            receptive_field_size *= dim
        fan_in = shape[-2] * receptive_field_size
        fan_out = shape[-1] * receptive_field_size
    return int(fan_in), int(fan_out)


@ks.utils.register_keras_serializable(package='kgcnn', name='glorot_orthogonal')
class GlorotOrthogonal(ks.initializers.Orthogonal):
    r"""Combining Glorot variance and Orthogonal initializer.

    Generate a weight matrix with variance according to Glorot initialization.
    Based on a random (semi-) orthogonal matrix neural networks
    are expected to learn better when features are de-correlated.

    This is stated by e.g.:

        * "Reducing over-fitting in deep networks by de-correlating representations" by M. Cogswell et al. (2016)
          `<https://arxiv.org/abs/1511.06068>`_ .
        * "Dropout: a simple way to prevent neural networks from over-fitting" by N. Srivastava et al. (2014)
          `<https://dl.acm.org/doi/10.5555/2627435.2670313>`_ .
        * "Exact solutions to the nonlinear dynamics of learning in deep linear neural networks"
          by A. M. Saxe et al. (2013) `<https://arxiv.org/abs/1312.6120>`_ .

    This implementation has been borrowed and slightly modified from `DimeNetPP <https://arxiv.org/abs/2011.14115>`__ .
    """

    def __init__(self, gain=1.0, seed=None, scale=1.0, mode='fan_avg'):
        super(GlorotOrthogonal, self).__init__(gain=gain, seed=seed)
        self.scale = scale
        self.mode = mode

    def __call__(self, shape, dtype="float32", **kwargs):
        weight_kernel = super(GlorotOrthogonal, self).__call__(shape, dtype=dtype, **kwargs)
        # Original implementation from DimeNet.
        # assert len(shape) == 2
        # W = self.orth_init(shape, dtype)
        # W *= tf.sqrt(self.scale / ((shape[0] + shape[1]) * tf.math.reduce_variance(W)))  # scale = 2.0
        # Adapted with mode and scale chosen by class. Default values should match original version, to be used
        # for DimeNet model implementation.
        scale = self.scale
        fan_in, fan_out = _compute_fans(shape)
        if self.mode == "fan_in":
            scale /= max(1., fan_in)
        elif self.mode == "fan_out":
            scale /= max(1., fan_out)
        else:
            scale /= max(1., (fan_in + fan_out) / 2.)
        stddev = ops.sqrt(scale/ops.var(weight_kernel))
        weight_kernel *= stddev
        return weight_kernel

    def get_config(self):
        """Get keras config."""
        config = super(GlorotOrthogonal, self).get_config()
        config.update({"scale": self.scale, "mode": self.mode})
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='he_orthogonal')
class HeOrthogonal(ks.initializers.Orthogonal):
    """Combining He variance and Orthogonal initializer.

    Generate a weight matrix with variance according to He initialization.
    Based on a random (semi-)orthogonal matrix neural networks are expected to learn better
    when features are de-correlated.

    This is stated by e.g.:

        * "Reducing over-fitting in deep networks by de-correlating representations" by M. Cogswell et al. (2016)
          `<https://arxiv.org/abs/1511.06068>`_ .
        * "Dropout: a simple way to prevent neural networks from over-fitting" by N. Srivastava et al. (2014)
           `<https://dl.acm.org/doi/10.5555/2627435.2670313>`_ .
        * "Exact solutions to the nonlinear dynamics of learning in deep linear neural networks"
          by A. M. Saxe et al. (2013) `<https://arxiv.org/abs/1312.6120>`_ .

    This implementation has been borrowed and slightly modified from `GemNet <https://arxiv.org/abs/2106.08903>`__ .
    """

    def __init__(self, gain=1.0, seed=None, scale=1.0, mode='fan_in'):
        super(HeOrthogonal, self).__init__(gain=gain, seed=seed)
        self.scale = scale
        self.mode = mode

    def __call__(self, shape, dtype="float32", **kwargs):
        weight_kernel = super(HeOrthogonal, self).__call__(shape, dtype=dtype, **kwargs)

        # Original reference implementation was designed for kernel rank={2,3}.
        # fan_in = shape[0]
        # if len(shape) == 3:
        #     fan_in = fan_in * shape[1]
        # Tried to generalize with keras _compute_fans that is extends to convolutional kernels.
        # Although, not really meaningful in standard GNN applications.
        # Optionally use other scales.
        scale = self.scale
        fan_in, fan_out = _compute_fans(shape)
        if self.mode == "fan_in":
            scale /= max(1., fan_in)
        elif self.mode == "fan_out":
            scale /= max(1., fan_out)
        else:
            scale /= max(1., (fan_in + fan_out) / 2.)

        weight_kernel = self._standardize(weight_kernel, shape)

        # Original reference implementation with 1/fan_in changed to scale=scale/fan_in
        # W *= tf.sqrt(1 / fan_in)  # variance decrease is addressed in the dense layers
        weight_kernel *= ops.sqrt(scale)

        return weight_kernel

    @staticmethod
    def _standardize(kernel, shape):
        r"""Standardize kernel over `fan_in` dimensions.

        Args:
            kernel: Kernel variable.
            shape: Shape of the kernel.
        """
        # Original doc string: Makes sure that N*Var(W) = 1 and E[W] = 0
        # From original implementation as comments.
        # eps = 1e-6
        eps = ks.backend.epsilon()
        # if len(shape) == 3:
        #     axis = [0, 1]  # last dimension is output dimension
        if len(shape) == 0:
            # Constant does not really have variance.
            # Moreover, Orthogonal initializer should throw error.
            return kernel
        if len(shape) >= 3:
            axis = [i for i in range(len(shape)-1)]
        else:
            axis = 0
        mean = ops.mean(kernel, axis=axis, keepdims=True)
        var = ops.var(kernel, axis=axis, keepdims=True)
        kernel = (kernel - mean) / ops.sqrt(var + eps)
        return kernel

    def get_config(self):
        """Get keras config."""
        config = super(HeOrthogonal, self).get_config()
        config.update({"scale": self.scale, "mode": self.mode})
        return config



import keras as ks
from typing import Union
from kgcnn.layers.pooling import PoolingNodes
import numpy as np
from keras import ops


class StandardLabelScaler(ks.layers.Layer):  # noqa

    def __init__(self, scaling_shape: tuple = None, dtype_scale: str = "float64", trainable: bool = False,
                 name="StandardLabelScaler", **kwargs):
        r"""Initialize layer instance of :obj:`StandardLabelScaler` .

        Args:
            scaling_shape (tuple): Shape
        """
        super(StandardLabelScaler, self).__init__(**kwargs)
        self._scaling_shape = scaling_shape
        self.name = name
        self._weights_trainable = trainable
        self.dtype_scale = dtype_scale
        self.extensive = False

        if self._scaling_shape is not None:
            self._add_weights_for_scaling()

    def _add_weights_for_scaling(self):
        self.scale_ = self.add_weight(
            shape=self._scaling_shape,
            initializer="ones", trainable=self._weights_trainable, dtype=self.dtype_scale
        )
        self.mean_ = self.add_weight(
            shape=self._scaling_shape,
            initializer="zeros", trainable=self._weights_trainable, dtype=self.dtype_scale
        )

    def build(self, input_shape):
        if self._scaling_shape is None:
            if input_shape is None:
                raise ValueError("Can not build scale and mean weights if `input_shape` not known.")
            self._scaling_shape = tuple([1 if i is None else i for i in input_shape])
            self._add_weights_for_scaling()
        self.built = True

    def compute_output_shape(self, input_shape):
        return input_shape

    def call(self, inputs, **kwargs):
        return ops.cast(inputs, dtype=self.dtype_scale)*self.scale_ + self.mean_

    def get_config(self):
        config = super(StandardLabelScaler, self).get_config()
        config.update({})
        return config

    def set_scale(self, scaler):
        self.set_weights([scaler.get_scaling(), scaler.get_mean_shift()])


class ExtensiveMolecularLabelScaler(ks.layers.Layer):  # noqa

    max_atomic_number = 95

    def __init__(self, scaling_shape: tuple = None, dtype_scale: str = "float64", trainable: bool = False,
                 name="ExtensiveMolecularLabelScaler", **kwargs):
        r"""Initialize layer instance of :obj:`StandardLabelScaler` .

        Args:
            scaling_shape (tuple): Shape
        """
        super(ExtensiveMolecularLabelScaler, self).__init__(**kwargs)
        self._scaling_shape = scaling_shape
        self.name = name
        self._weights_trainable = trainable
        self.dtype_scale = dtype_scale
        self.extensive = True
        self.layer_pool = PoolingNodes(pooling_method="scatter_sum")

        self._fit_atom_selection_mask = self.add_weight(
            shape=(self.max_atomic_number, ), trainable=False, dtype="bool", initializer="zeros")

        if self._scaling_shape is not None:
            self._add_weights_for_scaling()

    def _add_weights_for_scaling(self):
        self.scale_ = self.add_weight(
            shape=self._scaling_shape,
            initializer="ones", trainable=self._weights_trainable, dtype=self.dtype_scale
        )
        self.ridge_kernel_ = self.add_weight(
            shape=tuple([self.max_atomic_number] + list(self._scaling_shape[1:])),
            initializer="zeros", trainable=self._weights_trainable, dtype=self.dtype_scale
        )

    def build(self, input_shape):
        if self._scaling_shape is None:
            if input_shape is None:
                raise ValueError("Can not build scale and mean weights if `input_shape` and `scaling_shape` not known.")
            self._scaling_shape = tuple([1 if i is None else i for i in input_shape[0]])
            self._add_weights_for_scaling()

        self.built = True

    def compute_output_shape(self, input_shape):
        return input_shape[0]

    def call(self, inputs, **kwargs):
        graph, nodes, batch_id = inputs
        energy_per_node = ops.take(self.ridge_kernel_, nodes, axis=0)
        extensive_energies = self.layer_pool([graph, energy_per_node, batch_id])
        return ops.cast(graph, dtype=self.dtype_scale)*self.scale_ + extensive_energies

    def get_config(self):
        config = super(ExtensiveMolecularLabelScaler, self).get_config()
        config.update({})
        return config

    def set_scale(self, scaler):
        ridge_kernel = np.transpose(np.array(scaler.ridge.coef_))
        pos = np.sort(np.array(scaler._fit_atom_selection))
        mask = np.array(scaler._fit_atom_selection_mask)
        shape = tuple([int(self.max_atomic_number)] + list(ridge_kernel.shape[1:]))
        layer_kernel = np.zeros(shape)
        layer_kernel[pos] = ridge_kernel
        layer_kernel[0] = 0.  # Make sure 0 is always 0.
        self.set_weights([mask, scaler.get_scaling(), layer_kernel])


class QMGraphLabelScaler(ks.layers.Layer):  # noqa

    max_atomic_number = 95

    def __init__(self, scaler_list: list = None, name="QMGraphLabelScaler", **kwargs):
        r"""Initialize layer instance of :obj:`StandardLabelScaler` .

        Args:
            scaler_list (list): List of scaler
        """
        super(QMGraphLabelScaler, self).__init__(**kwargs)
        self._scaler_list = scaler_list
        self.name = name
        self.extensive = True

    def build(self, input_shape):
        for scaler in self._scaler_list:
            if scaler.extensive:
                scaler.build(input_shape)
            else:
                scaler.build(input_shape[0])
        self.built = True

    def compute_output_shape(self, input_shape):
        return input_shape[0]

    def call(self, inputs, **kwargs):
        return inputs

    def get_config(self):
        config = super(QMGraphLabelScaler, self).get_config()
        config.update({})
        return config

    def set_scale(self, scaler):
        for s in self._scaler_list:
            s.set_scale(scaler)


def get(scale_name: str):
    scaler_reference = {
        "StandardLabelScaler": StandardLabelScaler,
        "ExtensiveMolecularLabelScaler": ExtensiveMolecularLabelScaler
    }
    return scaler_reference[scale_name]


import keras as ks
from keras import ops
from keras.losses import Loss
from keras.losses import mean_absolute_error
import keras.saving
from kgcnn.ops.core import decompose_ragged_tensor
# from kgcnn.ops.scatter import scatter_reduce_mean


@ks.saving.register_keras_serializable(package='kgcnn', name='MeanAbsoluteError')
class MeanAbsoluteError(Loss):

    def __init__(self, reduction="sum_over_batch_size", name="mean_absolute_error",
                 padded_disjoint: bool = False, dtype=None):
        super(MeanAbsoluteError, self).__init__(reduction=reduction, name=name, dtype=dtype)
        self.padded_disjoint = padded_disjoint

    def call(self, y_true, y_pred):
        if self.padded_disjoint:
            y_true = y_true[1:]
            y_pred = y_pred[1:]
        out = mean_absolute_error(y_true, y_pred)
        return out

    def get_config(self):
        config = super(MeanAbsoluteError, self).get_config()
        config.update({"padded_disjoint": self.padded_disjoint})
        return config


@ks.saving.register_keras_serializable(package='kgcnn', name='ForceMeanAbsoluteError')
class ForceMeanAbsoluteError(Loss):

    def __init__(self, reduction="sum_over_batch_size", name="force_mean_absolute_error",
                 squeeze_states: bool = True, find_padded_atoms: bool = True, dtype=None):
        super(ForceMeanAbsoluteError, self).__init__(reduction=reduction, name=name, dtype=dtype)
        self.squeeze_states = squeeze_states
        self.find_padded_atoms = find_padded_atoms

    def call(self, y_true, y_pred):
        # Shape: (batch, N, 3, S)
        if self.find_padded_atoms:
            check_nonzero = ops.cast(ops.logical_not(
                ops.all(ops.isclose(y_true, ops.convert_to_tensor(0., dtype=y_true.dtype)), axis=2)), dtype="int32")
            row_count = ops.sum(check_nonzero, axis=1)
            row_count = ops.where(row_count < 1, 1, row_count)  # Prevent divide by 0.
            norm = 1/ops.cast(row_count, dtype=y_true.dtype)
        else:
            norm = 1/ops.shape(y_true)[1]

        diff = ops.abs(y_true-y_pred)
        out = ops.sum(ops.mean(diff, axis=2), axis=1)*norm
        if not self.squeeze_states:
            out = ops.mean(out, axis=-1)
        return out

    def get_config(self):
        config = super(ForceMeanAbsoluteError, self).get_config()
        config.update({"find_padded_atoms": self.find_padded_atoms, "squeeze_states": self.squeeze_states})
        return config


@ks.saving.register_keras_serializable(package='kgcnn', name='DisjointForceMeanAbsoluteError')
class DisjointForceMeanAbsoluteError(Loss):
    """This is an experimental class for force loss of disjoint output."""

    # Ideally we want to pass batch IDs and energy shape to the loss to do scatter mean.
    # However, passing as attribute similar to mask, has not been working yet.

    def __init__(self, reduction="sum_over_batch_size", name="force_mean_absolute_error",
                 squeeze_states: bool = True, padded_disjoint: bool = False, dtype=None):
        super(DisjointForceMeanAbsoluteError, self).__init__(reduction=reduction, name=name, dtype=dtype)
        self.squeeze_states = squeeze_states
        self.padded_disjoint = padded_disjoint

    def call(self, y_true, y_pred):
        # Shape: ([N], 3, S)

        if self.padded_disjoint:
            # Mask Shape: ([N, S])
            mask = ops.logical_not(ops.all(ops.isclose(y_true, ops.convert_to_tensor(0., dtype=y_true.dtype)), axis=1))
            y_pred = y_pred * ops.cast(ops.expand_dims(mask, axis=1), dtype=y_pred.dtype)
            row_count = ops.sum(ops.cast(mask, dtype="int32"), axis=0)
            row_count = ops.where(row_count < 1, 1, row_count)  # Prevent divide by 0.
            # roq count shape ([S])
            norm = 1 / ops.cast(row_count, dtype=y_true.dtype)
        else:
            norm = 1/ops.shape(y_true)[0]

        diff = ops.abs(y_true-y_pred)
        out = ops.mean(diff, axis=1)
        out = ops.sum(out, axis=0)*norm

        if not self.squeeze_states:
            out = ops.mean(out, axis=-1)

        return out

    def get_config(self):
        config = super(DisjointForceMeanAbsoluteError, self).get_config()
        config.update({"find_padded_atoms": self.find_padded_atoms, "squeeze_states": self.squeeze_states})
        return config


@ks.saving.register_keras_serializable(package='kgcnn', name='BinaryCrossentropyNoNaN')
class BinaryCrossentropyNoNaN(ks.losses.BinaryCrossentropy):

    def __init__(self, *args, **kwargs):
        super(BinaryCrossentropyNoNaN, self).__init__(*args, **kwargs)

    def call(self, y_true, y_pred):
        is_nan = ops.isnan(y_true)
        y_pred = ops.where(is_nan, ops.zeros_like(y_pred), y_pred)
        y_true = ops.where(is_nan, ops.zeros_like(y_true), y_true)
        return super(BinaryCrossentropyNoNaN, self).call(y_true, y_pred)


@ks.saving.register_keras_serializable(package='kgcnn', name='RaggedValuesMeanAbsoluteError')
class RaggedValuesMeanAbsoluteError(Loss):

    def __init__(self, reduction="sum_over_batch_size", name="mean_absolute_error", dtype=None):
        super(RaggedValuesMeanAbsoluteError, self).__init__(reduction=reduction, name=name, dtype=dtype)

    def call(self, y_true, y_pred):
        y_true_values = decompose_ragged_tensor(y_true)[0]
        y_pred_values = decompose_ragged_tensor(y_pred)[0]
        return mean_absolute_error(y_true_values, y_pred_values)

    def get_config(self):
        config = super(RaggedValuesMeanAbsoluteError, self).get_config()
        return config



from keras.layers import Add, Subtract, Concatenate, Dense
from kgcnn.layers.geom import NodePosition, NodeDistanceEuclidean, BesselBasisLayer, EdgeAngle, ShiftPeriodicLattice, \
    SphericalBasisLayer
from kgcnn.layers.gather import GatherNodes
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.layers.mlp import MLP
from ._layers import DimNetInteractionPPBlock, EmbeddingDimeBlock, DimNetOutputBlock


def model_disjoint(
        inputs,
        use_node_embedding,
        input_node_embedding: dict = None,
        emb_size: int = None,
        out_emb_size: int = None,
        int_emb_size: int = None,
        basis_emb_size: int = None,
        num_blocks: int = None,
        num_spherical: int = None,
        num_radial: int = None,
        cutoff: float = None,
        envelope_exponent: int = None,
        num_before_skip: int = None,
        num_after_skip: int = None,
        num_dense_output: int = None,
        num_targets: int = None,
        activation: str = None,
        extensive: bool = None,
        output_init: str = None,
        use_output_mlp: bool = None,
        output_embedding: str = None,
        output_mlp: dict = None
):
    n, x, edi, adi, batch_id_node, count_nodes = inputs

    # Atom embedding
    if use_node_embedding:
        n = EmbeddingDimeBlock(**input_node_embedding)(n)

    # Calculate distances
    pos1, pos2 = NodePosition()([x, edi])
    d = NodeDistanceEuclidean()([pos1, pos2])
    rbf = BesselBasisLayer(num_radial=num_radial, cutoff=cutoff, envelope_exponent=envelope_exponent)(d)

    # Calculate angles
    v12 = Subtract()([pos1, pos2])
    a = EdgeAngle()([v12, adi])
    sbf = SphericalBasisLayer(num_spherical=num_spherical, num_radial=num_radial, cutoff=cutoff,
                              envelope_exponent=envelope_exponent)([d, a, adi])

    # Embedding block
    rbf_emb = Dense(emb_size, use_bias=True, activation=activation,
                    kernel_initializer="kgcnn>glorot_orthogonal")(rbf)
    n_pairs = GatherNodes()([n, edi])
    x = Concatenate(axis=-1)([n_pairs, rbf_emb])
    x = Dense(emb_size, use_bias=True, activation=activation, kernel_initializer="kgcnn>glorot_orthogonal")(x)
    ps = DimNetOutputBlock(emb_size, out_emb_size, num_dense_output, num_targets=num_targets,
                           output_kernel_initializer=output_init)([n, x, rbf, edi])

    # Interaction blocks
    add_xp = Add()
    for i in range(num_blocks):
        x = DimNetInteractionPPBlock(emb_size, int_emb_size, basis_emb_size, num_before_skip, num_after_skip)(
            [x, rbf, sbf, adi])

        p_update = DimNetOutputBlock(emb_size, out_emb_size, num_dense_output, num_targets=num_targets,
                                     output_kernel_initializer=output_init)([n, x, rbf, edi])
        ps = add_xp([ps, p_update])

    if extensive:
        out = PoolingNodes(pooling_method="sum")([count_nodes, ps, batch_id_node])
    else:
        out = PoolingNodes(pooling_method="mean")([count_nodes, ps, batch_id_node])

    if use_output_mlp:
        out = MLP(**output_mlp)(out)

    if output_embedding != "graph":
        raise ValueError("Unsupported output embedding for mode `DimeNetPP`. ")

    return out


def model_disjoint_crystal(
        inputs,
        use_node_embedding,
        input_node_embedding: dict = None,
        emb_size: int = None,
        out_emb_size: int = None,
        int_emb_size: int = None,
        basis_emb_size: int = None,
        num_blocks: int = None,
        num_spherical: int = None,
        num_radial: int = None,
        cutoff: float = None,
        envelope_exponent: int = None,
        num_before_skip: int = None,
        num_after_skip: int = None,
        num_dense_output: int = None,
        num_targets: int = None,
        activation: str = None,
        extensive: bool = None,
        output_init: str = None,
        use_output_mlp: bool = None,
        output_embedding: str = None,
        output_mlp: dict = None
    ):

    n, x, edi, adi, edge_image, lattice, batch_id_node, batch_id_edge, count_nodes = inputs

    # Atom embedding
    if use_node_embedding:
        n = EmbeddingDimeBlock(**input_node_embedding)(n)

    # Calculate distances
    pos1, pos2 = NodePosition()([x, edi])
    pos2 = ShiftPeriodicLattice()([pos2, edge_image, lattice, batch_id_edge])
    d = NodeDistanceEuclidean()([pos1, pos2])
    rbf = BesselBasisLayer(num_radial=num_radial, cutoff=cutoff, envelope_exponent=envelope_exponent)(d)

    # Calculate angles
    v12 = Subtract()([pos1, pos2])
    a = EdgeAngle()([v12, adi])
    sbf = SphericalBasisLayer(num_spherical=num_spherical, num_radial=num_radial, cutoff=cutoff,
                              envelope_exponent=envelope_exponent)([d, a, adi])

    # Embedding block
    rbf_emb = Dense(emb_size, use_bias=True, activation=activation,
                    kernel_initializer="kgcnn>glorot_orthogonal")(rbf)
    n_pairs = GatherNodes()([n, edi])
    x = Concatenate(axis=-1)([n_pairs, rbf_emb])
    x = Dense(emb_size, use_bias=True, activation=activation, kernel_initializer="kgcnn>glorot_orthogonal")(x)
    ps = DimNetOutputBlock(emb_size, out_emb_size, num_dense_output, num_targets=num_targets,
                           output_kernel_initializer=output_init)([n, x, rbf, edi])

    # Interaction blocks
    add_xp = Add()
    for i in range(num_blocks):
        x = DimNetInteractionPPBlock(emb_size, int_emb_size, basis_emb_size, num_before_skip, num_after_skip)(
            [x, rbf, sbf, adi])
        p_update = DimNetOutputBlock(emb_size, out_emb_size, num_dense_output, num_targets=num_targets,
                                     output_kernel_initializer=output_init)([n, x, rbf, edi])
        ps = add_xp([ps, p_update])

    if extensive:
        out = PoolingNodes(pooling_method="sum")([count_nodes, ps, batch_id_node])
    else:
        out = PoolingNodes(pooling_method="mean")([count_nodes, ps, batch_id_node])

    if use_output_mlp:
        out = MLP(**output_mlp)(out)

    if output_embedding != "graph":
        raise ValueError("Unsupported output embedding for mode `DimeNetPP`. ")

    return out



import numpy as np
import unittest
from kgcnn.metrics.metrics import ScaledForceMeanAbsoluteError


class TestScaledForceMeanAbsoluteError(unittest.TestCase):

    def test_correctness(self):

        m = ScaledForceMeanAbsoluteError()

        y_true = np.array([[[1.0, 1.0, 1.0], [2.0, 2.0, 2.0], [3.0, 3.0, 3.0]],
                           [[1.0, 1.0, 1.0], [2.0, 2.0, 2.0], [0.0, 0.0, 0.0]]])
        y_pred = np.array([[[1.0, 1.0, 1.0], [2.0, 2.0, 2.0], [3.0, 3.0, 3.0]],
                           [[2.0, 2.0, 2.0], [2.0, 2.0, 2.0], [0.0, 0.0, 0.0]]])
        m.update_state(y_true=y_true, y_pred=y_pred)
        result = np.array(m.result())
        print(result)
        # self.assertTrue(np.max(np.abs(result - expected_result)) < 1e-6)


if __name__ == "__main__":

    TestScaledForceMeanAbsoluteError().test_correctness()
    print("Tests passed.")


import keras as ks
from ._model import MEGAN
from kgcnn.models.utils import update_model_kwargs
from kgcnn.layers.modules import Input
from keras.backend import backend as backend_to_use
from kgcnn.models.casting import (template_cast_output, template_cast_list_input,
                                  template_cast_list_input_docs, template_cast_output_docs)
import kgcnn.ops.activ

# Keep track of model version from commit date in literature.
# To be updated if model is changed in a significant way.
__model_version__ = "2023-12-08"

# Supported backends
__kgcnn_model_backend_supported__ = ["tensorflow", "torch", "jax"]
if backend_to_use() not in __kgcnn_model_backend_supported__:
    raise NotImplementedError("Backend '%s' for model 'MEGAN' is not supported." % backend_to_use())


# Implementation of INorp in `tf.keras` from paper:
# 'Interaction Networks for Learning about Objects, Relations and Physics'
# by Peter W. Battaglia, Razvan Pascanu, Matthew Lai, Danilo Rezende, Koray Kavukcuoglu
# http://papers.nips.cc/paper/6417-interaction-networks-for-learning-about-objects-relations-and-physics
# https://arxiv.org/abs/1612.00222
# https://github.com/higgsfield/interaction_network_pytorch


model_default = {
    "name": "MEGAN",
    "inputs": [
        {'shape': (None, 128), 'name': "node_attributes", 'dtype': 'float32'},
        {'shape': (None, 64), 'name': "edge_attributes", 'dtype': 'float32'},
        {'shape': (None, 2), 'name': "edge_indices", 'dtype': 'int64'},
        {'shape': (1,), 'name': "graph_labels", 'dtype': 'float32'},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_edges", "dtype": "int64"}
    ],
    "input_tensor_type": "padded",
    'input_embedding': None,
    "cast_disjoint_kwargs": {},
    "units": [128],
    "activation": {"class_name": "function", "config": "kgcnn>leaky_relu2"},
    "use_bias":  True,
    "dropout_rate":  0.0,
    "use_edge_features":  True,
    "input_node_embedding":  None,
    # node/edge importance related arguments
    "importance_units":  [],
    "importance_channels":  2,
    "importance_activation":  "sigmoid",  # do not change
    "importance_dropout_rate":  0.0,  # do not change
    "importance_factor":  0.0,
    "importance_multiplier":  10.0,
    "sparsity_factor":  0.0,
    "concat_heads":  True,
    # mlp tail end related arguments
    "final_units":  [1],
    "final_dropout_rate":  0.0,
    "final_activation":  'linear',
    "final_pooling":  'sum',
    "regression_limits":  None,
    "regression_reference": None,
    "return_importances": True,
    "output_tensor_type": "padded",
    'output_embedding': 'graph',
}


@update_model_kwargs(model_default, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_model(inputs: list = None,
               name: str = None,
               input_tensor_type: str = None,
               cast_disjoint_kwargs: dict = None,
               units: list = None,
               activation: str = None,
               use_bias: bool = None,
               dropout_rate: float = None,
               use_edge_features: bool = None,
               input_embedding: dict = None,  # deprecated
               input_node_embedding: dict = None,
               # node/edge importance related arguments
               importance_units: list = None,
               importance_channels: int = None,
               importance_activation: str = None,  # do not change
               importance_dropout_rate: float = None,  # do not change
               importance_factor: float = None,
               importance_multiplier: float = None,
               sparsity_factor: float = None,
               concat_heads: bool = None,
               # mlp tail end related arguments
               final_units: list = None,
               final_dropout_rate: float = None,
               final_activation: str = None,
               final_pooling: str = None,
               regression_limits: tuple = None,
               regression_reference: float = None,
               return_importances: bool = True,
               output_embedding: str = None,
               output_tensor_type: str = None
               ):
    r"""Functional model definition of MEGAN. Please check documentation of :obj:`kgcnn.literature.MEGAN` .

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are  :obj:`[nodes, edges, edge_indices, graph_labels...]`
    with '...' indicating mask or ID tensors following the template below.
    Graph labels are used to generate explanations but not to influence model output.

    %s

    **Model outputs**:
    The standard output template:

    %s

    Args:
        name: Name of the model.
        inputs (list): List of dictionaries unpacked in :obj:`keras.layers.Input`. Order must match model definition.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layer.
        units: A list of ints where each element configures an additional attention layer. The numeric
                value determines the number of hidden units to be used in the attention heads of that layer
        activation: The activation function to be used within the attention layers of the network
        use_bias: Whether the layers of the network should use bias weights at all
        dropout_rate: The dropout rate to be applied after *each* of the attention layers of the network.
        input_node_embedding: Dictionary of embedding kwargs for input embedding layer.
        use_edge_features: Whether edge features should be used. Generally the network supports the
            usage of edge features, but if the input data does not contain edge features, this should be
            set to False.
        importance_units: A list of ints where each element configures another dense layer in the
            subnetwork that produces the node importance tensor from the main node embeddings. The
            numeric value determines the number of hidden units in that layer.
        importance_channels: The int number of explanation channels to be produced by the network. This
            is the value referred to as "K". Note that this will also determine the number of attention
            heads used within the attention subnetwork.
        importance_factor: The weight of the explanation-only train step. If this is set to exactly
            zero then the explanation train step will not be executed at all (less computationally
            expensive)
        importance_multiplier: An additional hyperparameter of the explanation-only train step. This
            is essentially the scaling factor that is applied to the values of the dataset such that
            the target values can reasonably be approximated by a sum of [0, 1] importance values.
        sparsity_factor: The coefficient for the sparsity regularization of the node importance
            tensor.
        concat_heads: Whether to concat the heads of the attention subnetwork. The default is True. In
            that case the output of each individual attention head is concatenated and the concatenated
            vector is then used as the input of the next attention layer's heads. If this is False, the
            vectors are average pooled instead.
        final_units: A list of ints where each element configures another dense layer in the MLP
            at the tail end of the network. The numeric value determines the number of the hidden units
            in that layer. Note that the final element in this list has to be the same as the dimension
            to be expected for the samples of the training dataset!
        final_dropout_rate: The dropout rate to be applied after *every* layer of the final MLP.
        final_activation: The activation to be applied at the very last layer of the MLP to produce the
            actual output of the network.
        final_pooling: The pooling method to be used during the global pooling phase in the network.
        regression_limits: A tuple where the first value is the lower limit for the expected value range
            of the regression task and teh second value the upper limit.
        regression_reference: A reference value which is inside the range of expected values (best if
            it was in the middle, but does not have to). Choosing different references will result
            in different explanations.
        return_importances: Whether the importance / explanation tensors should be returned as an output
            of the model. If this is True, the output of the model will be a 3-tuple:
            (output, node importances, edge importances), otherwise it is just the output itself
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".

    Returns:
        :obj:`keras.models.Model`
    """
    model_inputs = [Input(**x) for x in inputs]

    di_inputs = template_cast_list_input(
        model_inputs,
        input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        mask_assignment=[0, 1, 1, None],
        index_assignment=[None, None, 0, None]
    )

    n, ed, disjoint_indices, gs, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges = di_inputs

    # Wrapping disjoint model.
    out = MEGAN(
        units=units,
        activation=activation,
        use_bias=use_bias,
        dropout_rate=dropout_rate,
        use_edge_features=use_edge_features,
        input_node_embedding=input_node_embedding,
        importance_units=importance_units,
        importance_channels=importance_channels,
        importance_activation=importance_activation,  # do not change
        importance_dropout_rate=importance_dropout_rate,  # do not change
        importance_factor=importance_factor,
        importance_multiplier=importance_multiplier,
        sparsity_factor=sparsity_factor,
        concat_heads=concat_heads,
        final_units=final_units,
        final_dropout_rate=final_dropout_rate,
        final_activation=final_activation,
        final_pooling=final_pooling,
        regression_limits=regression_limits,
        regression_reference=regression_reference,
        return_importances=return_importances
    )([n, ed, disjoint_indices, gs, batch_id_node, count_nodes])

    # Output embedding choice
    out = template_cast_output(
        [out, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding, output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type, cast_disjoint_kwargs=cast_disjoint_kwargs,
    )

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)
    model.__kgcnn_model_version__ = __model_version__

    return model


make_model.__doc__ = make_model.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)



from ._make import make_model, model_default
from ._make import make_crystal_model, model_crystal_default


__all__ = [
    "make_model",
    "model_default",
    "make_crystal_model",
    "model_crystal_default"
]



import os
import numpy as np
from kgcnn.molecule.base import MolGraphInterface
from typing import Union, Callable, Dict
from kgcnn.data.qm import QMDataset
from kgcnn.molecule.io import parse_list_to_xyz_str, read_xyz_file, write_list_to_xyz_file


class ForceDataset(QMDataset):
    r"""This is a base class for Force datasets. Inherits all functionality from QMDataset.

    It generates graph properties from a xyz-file, which stores atomic coordinates.
    Additionally, loading multiple single xyz-files into one file is supported. The file names and labels are given
    by a CSV or table file. The table file must have one line of header with column names!

    .. code-block:: console

        ├── data_directory
            ├── file_directory
            │   ├── *.xyz
            │   ├── *.xyz
            │   └── ...
            ├── file_name.csv
            ├── file_name.xyz
            ├── file_name.sdf
            ├── file_name_force.xyz
            ├── ...
            └── dataset_name.kgcnn.pickle

    Additionally, forces xyz information can be read in with this class.
    """

    def __init__(self, data_directory: str = None, dataset_name: str = None, file_name: str = None,
                 verbose: int = 10, file_directory: str = None,
                 file_name_xyz: str = None,
                 file_name_mol: str = None,
                 file_name_force_xyz: str = None):
        r"""Default initialization. File information on the location of the dataset on disk should be provided here.

        Args:
            data_directory (str): Full path to directory of the dataset. Optional. Default is None.
            file_name (str): Filename for reading table '.csv' file into memory. Must be given!
                For example as '.csv' formatted file with QM labels such as energy, states, dipole etc.
                Moreover, the table file can contain a list of file names of individual '.xyz' files to collect.
                Files are expected to be in :obj:`file_directory`. Default is None.
            file_name_xyz (str): Filename of a single '.xyz' file. This file is generated when collecting single
                '.xyz' files in :obj:`file_directory` . If not specified, the name is generated based on
                :obj:`file_name` .
            file_name_mol (str): Filename of a single '.sdf' file. This file is generated from the single '.xyz'
                file. SDF generation does require proper geometries. If not specified, the name is generated based on
                :obj:`file_name` .
            file_directory (str): Name or relative path from :obj:`data_directory` to a directory containing sorted
                '.xyz' files. Only used if :obj:`file_name` is None. Default is None.
            dataset_name (str): Name of the dataset. Important for naming and saving files. Default is None.
            verbose (int): Logging level. Default is 10.
        """
        super(ForceDataset, self).__init__(data_directory=data_directory, dataset_name=dataset_name,
                                           file_name=file_name, verbose=verbose, file_directory=file_directory)
        self.label_units = None
        self.label_names = None
        self.file_name_xyz = file_name_xyz
        self.file_name_mol = file_name_mol
        self.file_name_force_xyz = file_name_force_xyz

    @property
    def file_path_force_xyz(self):
        """Try to determine a file name for the mol information to store."""
        self._verify_data_directory()
        if self.file_name_force_xyz is None:
            return os.path.join(self.data_directory, os.path.splitext(self.file_name)[0] + "_force.xyz")
        elif isinstance(self.file_name_force_xyz, (str, os.PathLike)):
            return os.path.join(self.data_directory, self.file_name_force_xyz)
        elif isinstance(self.file_name_force_xyz, (list, tuple)):
            return [os.path.join(self.data_directory, x) for x in self.file_name_force_xyz]
        else:
            raise TypeError("Wrong type for `file_name_force_xyz` : '%s'." % self.file_name_force_xyz)

    def prepare_data(self, overwrite: bool = False, file_column_name: str = None, file_column_name_force: str = None,
                     make_sdf: bool = False):
        r"""Pre-computation of molecular structure information in a sdf-file from a xyz-file or a folder of xyz-files.

        If there is no single xyz-file, it will be created with the information of a csv-file with the same name.

        Args:
            overwrite (bool): Overwrite existing database SDF file. Default is False.
            file_column_name (str): Name of the column in csv file with list of xyz-files located in file_directory.
                This is for the positions only.
            file_column_name_force (str, list): Column name of xyz files for forces in file directory.
            make_sdf (bool): Whether to try to make a sdf file from xyz information via `RDKit` and `OpenBabel`.

        Returns:
            self
        """
        super(ForceDataset, self).prepare_data(overwrite=overwrite, file_column_name=file_column_name,
                                               make_sdf=make_sdf)

        # Try collect single xyz files in directory
        file_path_forces = self.file_path_force_xyz
        if not isinstance(file_path_forces, (list, tuple)):
            file_path_forces = [file_path_forces]
        if not isinstance(file_column_name_force, (list, tuple)):
            file_column_name_force = [file_column_name_force]
        for f, c in zip(file_path_forces, file_column_name_force):
            if not os.path.exists(f):
                xyz_list = self.collect_files_in_file_directory(
                    file_column_name=c, table_file_path=None,
                    read_method_file=self.get_geom_from_xyz_file, update_counter=self._default_loop_update_info,
                    append_file_content=True, read_method_return_list=True
                )
                write_list_to_xyz_file(f, xyz_list)

        return self

    def read_in_memory(self,
                       label_column_name: Union[str, list] = None,
                       nodes: list = None,
                       edges: list = None,
                       graph: list = None,
                       encoder_nodes: dict = None,
                       encoder_edges: dict = None,
                       encoder_graph: dict = None,
                       add_hydrogen: bool = True,
                       sanitize: bool = False,
                       make_directed: bool = False,
                       compute_partial_charges: bool = False,
                       additional_callbacks: Dict[str, Callable[[MolGraphInterface, dict], None]] = None,
                       custom_transform: Callable[[MolGraphInterface], MolGraphInterface] = None):
        """Read geometric information into memory.

        Graph labels require a column specified by :obj:`label_column_name`.

        Args:
            label_column_name (str, list): Name of labels for columns in CSV file.
            nodes (list): A list of node attributes as string. In place of names also functions can be added.
            edges (list): A list of edge attributes as string. In place of names also functions can be added.
            graph (list): A list of graph attributes as string. In place of names also functions can be added.
            encoder_nodes (dict): A dictionary of callable encoder where the key matches the attribute.
            encoder_edges (dict): A dictionary of callable encoder where the key matches the attribute.
            encoder_graph (dict): A dictionary of callable encoder where the key matches the attribute.
            add_hydrogen (bool): Whether to keep hydrogen after reading the mol-information. Default is False.
            make_directed (bool): Whether to have directed or undirected bonds. Default is False.
            compute_partial_charges (str): Whether to compute partial charges, e.g. 'gasteiger'. Default is None.
            sanitize (bool): Whether to sanitize molecule. Default is False.
            additional_callbacks (dict): A dictionary whose keys are string attribute names which the elements of the
                dataset are supposed to have and the elements are callback function objects which implement how those
                attributes are derived from the :obj:`MolecularGraphRDKit` of the molecule in question or the
                row of the CSV file.
            custom_transform (Callable): Custom transformation function to modify the generated
                :obj:`MolecularGraphRDKit` before callbacks are carried out. The function must take a single
                :obj:`MolecularGraphRDKit` instance as argument and return a (new) :obj:`MolecularGraphRDKit` instance.

        Returns:
            self
        """
        super(ForceDataset, self).read_in_memory(
            label_column_name=label_column_name, nodes=nodes, edges=edges, graph=graph, encoder_nodes=encoder_nodes,
            encoder_edges=encoder_edges, encoder_graph=encoder_graph, add_hydrogen=add_hydrogen, sanitize=sanitize,
            make_directed=make_directed, compute_partial_charges=compute_partial_charges,
            additional_callbacks=additional_callbacks, custom_transform=custom_transform
        )
        file_path_forces = self.file_path_force_xyz
        if not isinstance(file_path_forces, (list, tuple)):
            file_path_forces = [file_path_forces]
        for i, x in enumerate(file_path_forces):
            self.read_in_memory_xyz(x, atomic_coordinates=str(os.path.basename(x)),
                                    # We do not want to change atomic information but only read xyz.
                                    atomic_number=None, atomic_symbol=None)
        return self



import keras as ks
from keras.layers import Layer
from keras import ops
from kgcnn.ops.core import repeat_static_length, decompose_ragged_tensor
from kgcnn.ops.scatter import scatter_reduce_sum
from kgcnn import __indices_axis__ as global_axis_indices


def _pad_left(t):
    return ops.pad(t, [[1, 0]] + [[0, 0] for _ in range(len(ops.shape(t)) - 1)])


def _cat_one(t):
    return ops.concatenate([ops.convert_to_tensor([1], dtype=t.dtype), t], axis=0)


class _CastBatchedDisjointBase(Layer):

    def __init__(self, reverse_indices: bool = False, dtype_batch: str = "int64", dtype_index=None,
                 padded_disjoint: bool = False, uses_mask: bool = False,
                 static_batched_node_output_shape: tuple = None,
                 static_batched_edge_output_shape: tuple = None,
                 remove_padded_disjoint_from_batched_output: bool = True,
                 **kwargs):
        r"""Initialize layer.

        Args:
            reverse_indices (bool): Whether to reverse index order. Default is False.
            dtype_batch (str): Dtype for batch ID tensor. Default is 'int64'.
            dtype_index (str): Dtype for index tensor. Default is None.
            padded_disjoint (bool): Whether to keep padding in disjoint representation. Default is False.
            uses_mask (bool): Whether the padding is marked by a boolean mask or by a length tensor, counting the
                non-padded nodes from index 0. Default is False.
            static_batched_node_output_shape (tuple): Statical output shape of nodes. Default is None.
            static_batched_edge_output_shape (tuple): Statical output shape of edges. Default is None.
            remove_padded_disjoint_from_batched_output (bool): Whether to remove the first element on batched output
                in case of padding.
        """
        super(_CastBatchedDisjointBase, self).__init__(**kwargs)
        self.reverse_indices = reverse_indices
        self.dtype_index = dtype_index
        self.dtype_batch = dtype_batch
        self.uses_mask = uses_mask
        self.padded_disjoint = padded_disjoint
        self.supports_jit = padded_disjoint
        self.static_batched_node_output_shape = static_batched_node_output_shape
        self.static_batched_edge_output_shape = static_batched_edge_output_shape
        self.remove_padded_disjoint_from_batched_output = remove_padded_disjoint_from_batched_output

    def get_config(self):
        """Get config dictionary for this layer."""
        config = super(_CastBatchedDisjointBase, self).get_config()
        config.update({"reverse_indices": self.reverse_indices, "dtype_batch": self.dtype_batch,
                       "dtype_index": self.dtype_index, "padded_disjoint": self.padded_disjoint,
                       "uses_mask": self.uses_mask,
                       "static_batched_node_output_shape": self.static_batched_node_output_shape,
                       "static_batched_edge_output_shape": self.static_batched_edge_output_shape,
                       "remove_padded_disjoint_from_batched_output": self.remove_padded_disjoint_from_batched_output
                       })
        return config


class CastBatchedIndicesToDisjoint(_CastBatchedDisjointBase):
    r"""Cast batched node and edge indices to a (single) disjoint graph representation of
    `Pytorch Geometric (PyG) <https://github.com/pyg-team/pytorch_geometric>`__ .
    For PyG a batch of graphs is represented by single graph which contains disjoint sub-graphs,
    and the batch information is passed as batch ID tensor: `graph_id_node` and `graph_id_edge` .

    Keras layers can pass unstacked tensors without batch dimension, however, for model input and output
    batched tensors is most natural to the framework. Therefore, this layer can cast to disjoint from padded
    input and also keep padding in disjoint representation for jax.

    For padded disjoint all padded nodes are assigned to a padded first empty graph, with single node and at least
    a single self-loop. This graph therefore does not interact with the actual graphs in the message passing.

    .. warning::

        However, for special operations such as :obj:`GraphBatchNormalization` the information of :obj:`padded_disjoint`
        must be separately provided, otherwise this will lead to unwanted behaviour.
    """

    def __init__(self,  **kwargs):
        super(CastBatchedIndicesToDisjoint, self).__init__(**kwargs)

    def build(self, input_shape):
        """Build layer."""
        # Not variables or sub-layers. Nothing to build.
        self.built = True

    def compute_output_spec(self, inputs_spec):
        """Compute output spec as possible."""
        output_shape = self.compute_output_shape([x.shape for x in inputs_spec])
        dtype_batch = inputs_spec[2].dtype if self.dtype_batch is None and not self.uses_mask else self.dtype_batch
        dtype_index = inputs_spec[1].dtype if self.dtype_index is None else self.dtype_index
        output_dtypes = [inputs_spec[0].dtype, dtype_index, dtype_batch, dtype_batch, dtype_batch, dtype_batch,
                         dtype_batch, dtype_batch]
        output_spec = [ks.KerasTensor(s, dtype=d) for s, d in zip(output_shape, output_dtypes)]
        return output_spec

    def compute_output_shape(self, input_shape):
        """Compute output shape as possible."""
        in_n, in_i, in_size_n, in_size_e = input_shape

        if not self.padded_disjoint:
            out_n = tuple([None] + list(in_n[2:]))
        else:
            out_n = tuple([in_n[0]*in_n[1]+1 if in_n[0] is not None and in_n[1] is not None else None] + list(in_n[2:]))

        out_gn = (None, ) if out_n[0] is None else out_n[:1]
        out_id_n = (None, ) if out_n[0] is None else out_n[:1]

        if not self.padded_disjoint:
            out_i = tuple([None] + list(in_i[2:]))
        else:
            out_i = tuple(
                [in_i[0] * in_i[1] + 1 if in_i[0] is not None and in_i[1] is not None else None] + list(in_i[2:]))

        out_ge = (None,) if out_i[0] is None else tuple([out_i[0]])
        out_id_e = (None,) if out_i[0] is None else tuple([out_i[0]])

        if global_axis_indices == 0:
            out_i = tuple(reversed(list(out_i)))

        batch_dim_n = in_size_n[0] if not self.uses_mask else in_n[0]
        batch_dim_e = in_size_e[0] if not self.uses_mask else in_i[0]
        if self.padded_disjoint:
            out_size_n = (batch_dim_n+1, ) if batch_dim_n is not None else (None, )
            out_size_e = (batch_dim_e+1, ) if batch_dim_e is not None else (None, )
        else:
            out_size_n, out_size_e = (batch_dim_n, ), (batch_dim_e, )

        return out_n, out_i, out_gn, out_ge, out_id_n, out_id_e, out_size_n, out_size_e

    def call(self, inputs: list, **kwargs):
        r"""Changes node and edge indices into a Pytorch Geometric (PyG) compatible tensor format.

        Args:
            inputs (list): List of `[nodes, edge_indices, nodes_in_batch/node_mask, edges_in_batch/edge_mask]` ,

                - nodes (Tensor): Node features are represented by a keras tensor of shape `(batch, N, F, ...)` ,
                  where N denotes the number of nodes.
                - edge_indices (Tensor): Edge index list have shape `(batch, M, 2)` with the indices of M directed
                  edges at last axis for each edge.
                - total_nodes (Tensor): Tensor of number of nodes for each graph of shape `(batch, )` .
                - total_edges (Tensor): Tensor of number of edges for each graph of shape `(batch, )` .

        Returns:
            list: `[node_attr, edge_index, graph_id_node, graph_id_edge, node_id, edge_id, nodes_count, edges_count]`

                - node_attr (Tensor): Represents node attributes or coordinates of shape `([N], F, ...)` ,
                - edge_index (Tensor): Represents the index table of shape `(2, [M])` for directed edges.
                - graph_id_node (Tensor): ID tensor of batch assignment in disjoint graph of shape `([N], )` .
                - graph_id_edge (Tensor): ID tensor of batch assignment in disjoint graph of shape `([M], )` .
                - nodes_id (Tensor): The ID-tensor to assign each node to its respective graph of shape `([N], )` .
                - edges_id (Tensor): The ID-tensor to assign each edge to its respective graph of shape `([M], )` .
                - nodes_count (Tensor): Tensor of number of nodes for each graph of shape `(batch, )` .
                - edges_count (Tensor): Tensor of number of edges for each graph of shape `(batch, )` .
        """
        nodes, edge_indices, node_pad, edge_pad = inputs

        # def make_mask_flatten(len_per_dim, target_shape):
        #     mask = ops.reshape(ops.repeat(
        #         ops.convert_to_tensor([[True, False]], dtype="bool"), ops.shape(len_per_dim)[0], axis=0), (-1,))
        #     mask = ops.repeat(mask, ops.reshape(ops.concatenate([ops.expand_dims(len_per_dim, axis=-1),
        #            ops.expand_dims(target_shape[1] - len_per_dim, axis=-1)], axis=-1), [-1]), axis=0)
        #     return mask

        if self.dtype_index is not None:
            edge_indices = ops.cast(edge_indices, dtype=self.dtype_index)

        if self.dtype_batch is None:
            if self.uses_mask:
                raise ValueError("Require `dtype_batch` for batch ID tensor when using boolean mask.")
            dtype_batch = node_pad.dtype
        else:
            dtype_batch = self.dtype_batch

        if not self.uses_mask:
            node_len = ops.cast(node_pad, dtype=dtype_batch)
            edge_len = ops.cast(edge_pad, dtype=dtype_batch)
            node_id = ops.repeat(ops.expand_dims(ops.arange(ops.shape(nodes)[1], dtype=dtype_batch), axis=0),
                                 ops.shape(node_len)[0], axis=0)
            edge_id = ops.repeat(ops.expand_dims(ops.arange(ops.shape(edge_indices)[1], dtype=dtype_batch), axis=0),
                                 ops.shape(edge_len)[0], axis=0)
            node_mask = node_id < ops.expand_dims(node_len, axis=-1)
            edge_mask = edge_id < ops.expand_dims(edge_len, axis=-1)
        else:
            node_mask = node_pad
            edge_mask = edge_pad
            node_len = ops.sum(ops.cast(node_mask, dtype=dtype_batch), axis=1)
            edge_len = ops.sum(ops.cast(edge_mask, dtype=dtype_batch), axis=1)
            node_id = ops.repeat(ops.expand_dims(ops.arange(ops.shape(nodes)[1], dtype=dtype_batch), axis=0),
                                 ops.shape(node_len)[0], axis=0)
            edge_id = ops.repeat(ops.expand_dims(ops.arange(ops.shape(edge_indices)[1], dtype=dtype_batch), axis=0),
                                 ops.shape(edge_len)[0], axis=0)

        if not self.padded_disjoint:
            edge_indices_flatten = edge_indices[edge_mask]
            nodes_flatten = nodes[node_mask]
            node_id = node_id[node_mask]
            edge_id = edge_id[edge_mask]
            node_splits = ops.pad(ops.cumsum(node_len), [[1, 0]])
            graph_id_node = ops.repeat(ops.arange(ops.shape(node_len)[0], dtype=dtype_batch), node_len)
            graph_id_edge = ops.repeat(ops.arange(ops.shape(edge_len)[0], dtype=dtype_batch), edge_len)
            # offset_edge_indices = ops.expand_dims(ops.repeat(node_splits[:-1], edge_len), axis=-1)
            offset_edge_indices = ops.expand_dims(ops.take(node_splits, graph_id_edge, axis=0), axis=-1)
            offset_edge_indices = ops.broadcast_to(offset_edge_indices, ops.shape(edge_indices_flatten))
            disjoint_indices = edge_indices_flatten + ops.cast(offset_edge_indices, edge_indices_flatten.dtype)
        else:
            nodes_flatten = ops.reshape(nodes, [-1] + list(ops.shape(nodes)[2:]))
            edge_indices_flatten = ops.reshape(edge_indices, [-1] + list(ops.shape(edge_indices)[2:]))
            node_len_flat = ops.repeat(ops.cast([ops.shape(nodes)[1]], dtype=dtype_batch), ops.shape(nodes)[0])
            edge_len_flat = ops.repeat(ops.cast([ops.shape(edge_indices)[1]], dtype=dtype_batch),
                                       ops.shape(edge_indices)[0])
            edge_mask_flatten = ops.reshape(edge_mask, [-1])
            node_mask_flatten = ops.reshape(node_mask, [-1])
            node_id = ops.reshape(node_id, [-1])
            edge_id = ops.reshape(edge_id, [-1])

            nodes_flatten = _pad_left(nodes_flatten)
            edge_indices_flatten = _pad_left(edge_indices_flatten)
            node_id = _pad_left(node_id)
            edge_id = _pad_left(edge_id)
            node_len_flat = _cat_one(node_len_flat)
            edge_len_flat = _cat_one(edge_len_flat)
            node_mask_flatten = _pad_left(node_mask_flatten)
            edge_mask_flatten = _pad_left(edge_mask_flatten)

            graph_id_node = repeat_static_length(
                ops.arange(ops.shape(node_len_flat)[0], dtype=dtype_batch), node_len_flat,
                total_repeat_length=ops.shape(nodes_flatten)[0])
            graph_id_edge = repeat_static_length(
                ops.arange(ops.shape(edge_len_flat)[0], dtype=dtype_batch), edge_len_flat,
                total_repeat_length=ops.shape(edge_indices_flatten)[0])
            graph_id_node = ops.where(node_mask_flatten, graph_id_node, 0)
            graph_id_edge = ops.where(edge_mask_flatten, graph_id_edge, 0)

            node_id = ops.where(node_mask_flatten, node_id, 0)
            edge_id = ops.where(edge_mask_flatten, edge_id, 0)

            node_splits = ops.pad(ops.cumsum(node_len_flat), [[1, 0]])
            # offset_edge_indices = repeat_static_length(node_splits[:-1], edge_len_flat, total_repeat_length=ops.shape(edge_indices_flatten)[0])
            offset_edge_indices = ops.take(node_splits, graph_id_edge, axis=0)
            offset_edge_indices = ops.expand_dims(offset_edge_indices, axis=-1)
            offset_edge_indices = ops.broadcast_to(offset_edge_indices, ops.shape(edge_indices_flatten))
            disjoint_indices = edge_indices_flatten + ops.cast(offset_edge_indices, edge_indices_flatten.dtype)
            edge_mask_flatten = ops.expand_dims(edge_mask_flatten, axis=-1)
            disjoint_indices = ops.where(edge_mask_flatten, disjoint_indices, 0)
            node_len = ops.concatenate([ops.sum(node_len_flat[1:] - node_len, axis=0, keepdims=True), node_len], axis=0)
            edge_len = ops.concatenate([ops.sum(edge_len_flat[1:] - edge_len, axis=0, keepdims=True), edge_len], axis=0)

        # Transpose edge indices.
        if global_axis_indices == 0:
            disjoint_indices = ops.transpose(disjoint_indices)

        if self.reverse_indices:
            disjoint_indices = ops.flip(disjoint_indices, axis=global_axis_indices)

        return [nodes_flatten, disjoint_indices, graph_id_node, graph_id_edge, node_id, edge_id, node_len, edge_len]


CastBatchedIndicesToDisjoint.__init__.__doc__ = _CastBatchedDisjointBase.__init__.__doc__


class CastBatchedAttributesToDisjoint(_CastBatchedDisjointBase):
    r"""Cast batched node and edge attributes to a (single) disjoint graph representation of
    `Pytorch Geometric (PyG) <https://github.com/pyg-team/pytorch_geometric>`__ .

    Only applies a casting of attribute tensors similar to :obj:`CastBatchedIndicesToDisjoint` but without any
    index adjustment. Produces the batch-ID tensor assignment.

    For padded disjoint all padded nodes are assigned to a padded first empty graph, with single node and at least
    a single self-loop. This graph therefore does not interact with the actual graphs in the message passing.

    .. warning::

        However, for special operations such as :obj:`GraphBatchNormalization` the information of :obj:`padded_disjoint`
        must be separately provided, otherwise this will lead to unwanted behaviour.
    """

    def __init__(self, **kwargs):
        super(CastBatchedAttributesToDisjoint, self).__init__(**kwargs)

    def build(self, input_shape):
        """Build layer."""
        self.built = True

    def compute_output_spec(self, inputs_spec):
        """Compute output spec as possible."""
        output_shape = self.compute_output_shape([x.shape for x in inputs_spec])
        dtype_batch = inputs_spec[1].dtype if self.dtype_batch is None and not self.uses_mask else self.dtype_batch
        output_dtypes = [inputs_spec[0].dtype, dtype_batch, dtype_batch, dtype_batch]
        output_spec = [ks.KerasTensor(s, dtype=d) for s, d in zip(output_shape, output_dtypes)]
        return output_spec

    def compute_output_shape(self, input_shape):
        """Compute output shape as possible."""
        in_n, in_size_n = input_shape
        if not self.padded_disjoint:
            out_n = tuple([None] + list(in_n[2:]))
        else:
            out_n = tuple(
                [in_n[0] * in_n[1] + 1 if in_n[0] is not None and in_n[1] is not None else None] + list(in_n[2:]))
        out_gn = (None,) if out_n[0] is None else out_n[:1]
        out_id_n = (None,) if out_n[0] is None else out_n[:1]

        batch_dim_n = in_size_n[0] if not self.uses_mask else in_n[0]
        if self.padded_disjoint:
            out_size_n = (batch_dim_n + 1,) if batch_dim_n is not None else (None,)
        else:
            out_size_n = (batch_dim_n,)
        return out_n, out_gn, out_id_n, out_size_n

    def call(self, inputs: list, **kwargs):
        r"""Changes node or edge tensors into a Pytorch Geometric (PyG) compatible tensor format.

        Args:
            inputs (list): List of `[attr, total_attr/mask_attr]` ,

                - attr (Tensor): Features are represented by a keras tensor of shape `(batch, N, F, ...)` ,
                  where N denotes the number of nodes or edges.
                - total_attr (Tensor): Tensor of lengths for each graph of shape `(batch, )` .

        Returns:
            list: `[attr, graph_id, item_id, item_counts]` .

                - attr (Tensor): Represents attributes or coordinates of shape `([N], F, ...)`
                - graph_id (Tensor): ID tensor of batch assignment in disjoint graph of shape `([N], )` .
                - item_id (Tensor): The ID-tensor to assign each node to its respective graph of shape `([N], )` .
                - item_counts (Tensor): Tensor of lengths for each graph of shape `(batch, )` .
        """
        nodes, node_pad = inputs

        if self.dtype_batch is None:
            if self.uses_mask:
                raise ValueError("Require `dtype_batch` for batch ID tensor when using boolean mask.")
            dtype_batch = node_pad.dtype
        else:
            dtype_batch = self.dtype_batch

        if not self.uses_mask:
            node_len = ops.cast(node_pad, dtype=dtype_batch)
            node_id = ops.repeat(ops.expand_dims(ops.arange(ops.shape(nodes)[1], dtype=dtype_batch), axis=0),
                             ops.shape(node_len)[0], axis=0)
            node_mask = node_id < ops.expand_dims(node_len, axis=-1)
        else:
            node_mask = node_pad
            node_len = ops.sum(ops.cast(node_mask, dtype=dtype_batch), axis=1)
            node_id = ops.repeat(ops.expand_dims(ops.arange(ops.shape(nodes)[1], dtype=dtype_batch), axis=0),
                                 ops.shape(node_len)[0], axis=0)

        if not self.padded_disjoint:
            nodes_flatten = nodes[node_mask]
            graph_id_node = ops.repeat(ops.arange(ops.shape(node_len)[0], dtype=dtype_batch), node_len)
            node_id = node_id[node_mask]
        else:
            nodes_flatten = ops.reshape(nodes, [-1] + list(ops.shape(nodes)[2:]))
            node_len_flat = ops.repeat(ops.cast([ops.shape(nodes)[1]], dtype=dtype_batch), ops.shape(nodes)[0])
            node_mask_flatten = ops.reshape(node_mask, [-1])
            node_id = ops.reshape(node_id, [-1])
            nodes_flatten = _pad_left(nodes_flatten)
            node_id = _pad_left(node_id)
            node_len_flat = _cat_one(node_len_flat)
            node_mask_flatten = _pad_left(node_mask_flatten)
            graph_id = repeat_static_length(
                ops.arange(ops.shape(node_len_flat)[0], dtype=self.dtype_batch), node_len_flat,
                total_repeat_length=ops.shape(nodes_flatten)[0])
            graph_id_node = ops.where(node_mask_flatten, graph_id, 0)
            node_id = ops.where(node_mask_flatten, node_id, 0)
            node_len = ops.concatenate([ops.sum(node_len_flat[1:] - node_len, axis=0, keepdims=True), node_len], axis=0)

        return [nodes_flatten, graph_id_node, node_id, node_len]


CastBatchedAttributesToDisjoint.__init__.__doc__ = _CastBatchedDisjointBase.__init__.__doc__


class CastDisjointToBatchedAttributes(_CastBatchedDisjointBase):
    r"""Cast batched node and edge attributes from a (single) disjoint graph representation of
    `Pytorch Geometric (PyG) <https://github.com/pyg-team/pytorch_geometric>`__ .

    Reconstructs batched tensor with the help of ID tensor information.
    """

    def __init__(self, static_output_shape: tuple = None, return_mask: bool = False, **kwargs):
        super(CastDisjointToBatchedAttributes, self).__init__(**kwargs)
        self.static_output_shape = static_output_shape
        self.return_mask = return_mask

    def build(self, input_shape):
        self.built = True

    def call(self, inputs: list, **kwargs):
        r"""Changes node or edge tensors into a Pytorch Geometric (PyG) compatible tensor format.

        Args:
            inputs (list): List of `[attr, graph_id_attr, (attr_id), attr_counts]` ,

                - attr (Tensor): Features are represented by a keras tensor of shape `([N], F, ...)` ,
                  where N denotes the number of nodes or edges.
                - graph_id_attr (Tensor): ID tensor of batch assignment in disjoint graph of shape `([N], )` .
                - attr_id (Tensor, optional): The ID-tensor to assign each node to its respective graph
                  of shape `([N], )` . For padded disjoint graphs this is required.
                - attr_counts (Tensor): Tensor of lengths for each graph of shape `(batch, )` .

        Returns:
            Tensor: Batched output tensor of node or edge attributes of shape `(batch, N, F, ...)` .
        """
        if len(inputs) == 4:
            attr, graph_id_attr, attr_id, attr_len = inputs
        else:
            attr, graph_id_attr, attr_len = inputs
            attr_id = None

        if self.static_output_shape is not None:
            target_shape = (ops.shape(attr_len)[0], self.static_output_shape[0])
        else:
            target_shape = (ops.shape(attr_len)[0], ops.cast(ops.amax(attr_len), dtype="int32"))
        out_mask = None

        if not self.padded_disjoint:
            if attr_id is None:
                attr_id = ops.arange(0, ops.shape(graph_id_attr)[0], dtype=graph_id_attr.dtype)
                attr_splits = ops.pad(ops.cumsum(attr_len), [[1, 0]])
                attr_id = attr_id - repeat_static_length(attr_splits[:1], attr_len, ops.shape(graph_id_attr)[0])
        else:
            if attr_id is None:
                # Required because padded graphs in the general case can have padded nodes inbetween batches.
                raise ValueError("Require sub-graph IDs in addition to batch IDs for padded disjoint graphs.")

        output_shape = tuple([target_shape[0] * target_shape[1]] + list(ops.shape(attr)[1:]))
        indices = graph_id_attr * ops.convert_to_tensor(target_shape[1], dtype=graph_id_attr.dtype) + ops.cast(
            attr_id, dtype=graph_id_attr.dtype)
        out = scatter_reduce_sum(indices, attr, output_shape)
        out = ops.reshape(out, list(target_shape[:2]) + list(ops.shape(attr)[1:]))
        if self.return_mask:
            output_mask_shape = output_shape[:1]
            out_mask = scatter_reduce_sum(indices, ops.ones(ops.shape(attr)[0], dtype="bool"), output_mask_shape)
            out_mask = ops.reshape(out_mask, list(target_shape[:2]))

        if self.padded_disjoint and self.remove_padded_disjoint_from_batched_output:
            out = out[1:]
            if self.return_mask:
                out_mask = out_mask[1:]

        if self.return_mask:
            return out, out_mask
        return out

    def get_config(self):
        """Get config dictionary for this layer."""
        config = super(_CastBatchedDisjointBase, self).get_config()
        config.update({"static_output_shape": self.static_output_shape, "return_mask": self.return_mask})
        return config


CastDisjointToBatchedAttributes.__init__.__doc__ = _CastBatchedDisjointBase.__init__.__doc__


class CastDisjointToBatchedGraphState(_CastBatchedDisjointBase):
    r"""Cast graph property tensor from disjoint graph representation of
    `Pytorch Geometric (PyG) <https://github.com/pyg-team/pytorch_geometric>`__ .

    The graph state is usually kept as batched tensor, except for padded disjoint representation, an empty zero valued
    graph is removed that represents all padded nodes.
    """

    def __init__(self, **kwargs):
        super(CastDisjointToBatchedGraphState, self).__init__(**kwargs)

    def build(self, input_shape):
        self.built = True

    def compute_output_shape(self, input_shape):
        if self.padded_disjoint and self.remove_padded_disjoint_from_batched_output:
            if input_shape[0] is not None:
                return tuple([input_shape[0] - 1] + list(input_shape[1:]))
        return input_shape

    def call(self, inputs: list, **kwargs):
        r"""Changes graph tensor from disjoint representation.

        Args:
            inputs (Tensor): Graph labels from a disjoint representation of shape `(batch, ...)` or
                `(batch+1, ...)` for padded disjoint.

        Returns:
            Tensor: Graph labels of shape `(batch, ...)` .
        """
        # Simply remove first graph.
        if self.padded_disjoint and self.remove_padded_disjoint_from_batched_output:
            return inputs[1:]
        return inputs


CastDisjointToBatchedGraphState.__init__.__doc__ = _CastBatchedDisjointBase.__init__.__doc__


class CastBatchedGraphStateToDisjoint(_CastBatchedDisjointBase):
    r"""Cast graph property tensor to disjoint graph representation of
    `Pytorch Geometric (PyG) <https://github.com/pyg-team/pytorch_geometric>`__ .

    The graph state is usually kept as batched tensor, except for padded disjoint representation, an empty zero valued
    graph is added to represent all padded nodes.
    """

    def __init__(self, **kwargs):
        super(CastBatchedGraphStateToDisjoint, self).__init__(**kwargs)

    def build(self, input_shape):
        self.built = True

    def compute_output_shape(self, input_shape):
        if self.padded_disjoint:
            if input_shape[0] is not None:
                return tuple([input_shape[0] + 1] + list(input_shape[1:]))
        return input_shape

    def compute_output_spec(self, input_spec):
        return ks.KerasTensor(self.compute_output_shape(input_spec.shape), input_spec.dtype)

    def call(self, inputs: list, **kwargs):
        r"""Changes graph tensor from disjoint representation.

        Args:
            inputs (Tensor): Graph labels from a disjoint representation of shape `(batch, ...)` .

        Returns:
            Tensor: Graph labels of shape `(batch, ...)` or `(batch+1, ...)` for padded disjoint.
        """
        if self.padded_disjoint:
            return _pad_left(inputs)
        return inputs


CastBatchedGraphStateToDisjoint.__init__.__doc__ = _CastBatchedDisjointBase.__init__.__doc__


class _CastRaggedToDisjointBase(Layer):

    def __init__(self, reverse_indices: bool = False, dtype_batch: str = "int64", dtype_index=None, **kwargs):
        r"""Initialize layer.

        Args:
            reverse_indices (bool): Whether to reverse index order. Default is False.
            dtype_batch (str): Dtype for batch ID tensor. Default is 'int64'.
            dtype_index (str): Dtype for index tensor. Default is None.
        """
        super(_CastRaggedToDisjointBase, self).__init__(**kwargs)
        self.reverse_indices = reverse_indices
        self.dtype_index = dtype_index
        self.dtype_batch = dtype_batch
        # self.supports_jit = False

    def get_config(self):
        """Get config dictionary for this layer."""
        config = super(_CastRaggedToDisjointBase, self).get_config()
        config.update({"reverse_indices": self.reverse_indices, "dtype_batch": self.dtype_batch,
                       "dtype_index": self.dtype_index})
        return config


class CastRaggedAttributesToDisjoint(_CastRaggedToDisjointBase):

    def __init__(self, **kwargs):
        super(CastRaggedAttributesToDisjoint, self).__init__(**kwargs)

    def compute_output_shape(self, input_shape):
        out_n = input_shape[1:]
        out_gn, out_id_n, out_size_n = out_n[:1], out_n[:1], input_shape[:1]
        return out_n, out_gn, out_id_n, out_size_n

    def compute_output_spec(self, inputs_spec):
        """Compute output spec as possible."""
        output_shape = self.compute_output_shape(inputs_spec.shape)
        dtype_batch = self.dtype_batch
        output_dtypes = [inputs_spec[0].dtype, dtype_batch, dtype_batch, dtype_batch]
        output_spec = [ks.KerasTensor(s, dtype=d) for s, d in zip(output_shape, output_dtypes)]
        return output_spec

    def build(self, input_shape):
        self.built = True

    def call(self, inputs, **kwargs):
        r"""Changes node or edge tensors into a Pytorch Geometric (PyG) compatible tensor format.

        Args:
            inputs (RaggedTensor): Attributes of shape `(batch, [None], F, ...)`

        Returns:
            list: `[attr, graph_id, item_id, item_counts]` .

                - attr (Tensor): Represents attributes or coordinates of shape `([N], F, ...)`
                - graph_id (Tensor): ID tensor of batch assignment in disjoint graph of shape `([N], )` .
                - item_id (Tensor): The ID-tensor to assign each node to its respective graph of shape `([N], )` .
                - item_counts (Tensor): Tensor of lengths for each graph of shape `(batch, )` .
        """
        return decompose_ragged_tensor(inputs, batch_dtype=self.dtype_batch)


CastRaggedAttributesToDisjoint.__init__.__doc__ = _CastRaggedToDisjointBase.__init__.__doc__


class CastRaggedIndicesToDisjoint(_CastRaggedToDisjointBase):

    def __init__(self, **kwargs):
        super(CastRaggedIndicesToDisjoint, self).__init__(**kwargs)

    def compute_output_spec(self, inputs_spec):
        """Compute output spec as possible."""
        output_shape = self.compute_output_shape([x.shape for x in inputs_spec])
        dtype_batch = self.dtype_batch
        dtype_index = inputs_spec[1].dtype if self.dtype_index is None else self.dtype_index
        output_dtypes = [inputs_spec[0].dtype, dtype_index, dtype_batch, dtype_batch, dtype_batch, dtype_batch,
                         dtype_batch, dtype_batch]
        output_spec = [ks.KerasTensor(s, dtype=d) for s, d in zip(output_shape, output_dtypes)]
        return output_spec

    def compute_output_shape(self, input_shape):
        """Compute output shape as possible."""
        in_n, in_i = input_shape

        out_n = tuple([None] + list(in_n[2:]))
        out_gn = (None, ) if out_n[0] is None else out_n[:1]
        out_id_n = (None, ) if out_n[0] is None else out_n[:1]

        out_i = tuple([None] + list(in_i[2:]))
        out_ge = (None,) if out_i[0] is None else tuple([out_i[0]])
        out_id_e = (None,) if out_i[0] is None else tuple([out_i[0]])

        if global_axis_indices == 0:
            out_i = tuple(reversed(list(out_i)))

        batch_dim_n = in_n[0]
        batch_dim_e = in_i[0]
        out_size_n, out_size_e = (batch_dim_n, ), (batch_dim_e, )

        return out_n, out_i, out_gn, out_ge, out_id_n, out_id_e, out_size_n, out_size_e

    def build(self, input_shape):
        self.built = True

    def call(self, inputs, **kwargs):
        r"""Changes node and edge indices into a Pytorch Geometric (PyG) compatible tensor format.

        Args:
            inputs (list): List of `[nodes, edge_indices]` ,

                - nodes (Tensor): Node features are represented by a keras tensor of shape `(batch, N, F, ...)` ,
                  where N denotes the number of nodes.
                - edge_indices (Tensor): Edge index list have shape `(batch, M, 2)` with the indices of M directed
                  edges at last axis for each edge.

        Returns:
            list: `[node_attr, edge_index, graph_id_node, graph_id_edge, node_id, edge_id, nodes_count, edges_count]`

                - node_attr (Tensor): Represents node attributes or coordinates of shape `([N], F, ...)` ,
                - edge_index (Tensor): Represents the index table of shape `(2, [M])` for directed edges.
                - graph_id_node (Tensor): ID tensor of batch assignment in disjoint graph of shape `([N], )` .
                - graph_id_edge (Tensor): ID tensor of batch assignment in disjoint graph of shape `([M], )` .
                - nodes_id (Tensor): The ID-tensor to assign each node to its respective graph of shape `([N], )` .
                - edges_id (Tensor): The ID-tensor to assign each edge to its respective graph of shape `([M], )` .
                - nodes_count (Tensor): Tensor of number of nodes for each graph of shape `(batch, )` .
                - edges_count (Tensor): Tensor of number of edges for each graph of shape `(batch, )` .
        """
        nodes, edge_indices = inputs
        nodes_flatten, graph_id_node, node_id, node_len = decompose_ragged_tensor(nodes, batch_dtype=self.dtype_batch)
        edge_indices_flatten, graph_id_edge, edge_id, edge_len = decompose_ragged_tensor(
            edge_indices, batch_dtype=self.dtype_batch)

        if self.dtype_index is not None:
            edge_indices_flatten = ops.cast(edge_indices_flatten, dtype=self.dtype_index)

        node_splits = ops.pad(ops.cumsum(node_len), [[1, 0]])
        offset_edge_indices = ops.expand_dims(ops.repeat(node_splits[:-1], edge_len), axis=-1)
        offset_edge_indices = ops.broadcast_to(offset_edge_indices, ops.shape(edge_indices_flatten))
        disjoint_indices = edge_indices_flatten + ops.cast(offset_edge_indices, edge_indices_flatten.dtype)

        # Transpose edge indices.
        if global_axis_indices == 0:
            disjoint_indices = ops.transpose(disjoint_indices)
        if self.reverse_indices:
            disjoint_indices = ops.flip(disjoint_indices, axis=global_axis_indices)

        return [nodes_flatten, disjoint_indices, graph_id_node, graph_id_edge, node_id, edge_id, node_len, edge_len]


CastRaggedIndicesToDisjoint.__init__.__doc__ = _CastRaggedToDisjointBase.__init__.__doc__


class CastDisjointToRaggedAttributes(_CastRaggedToDisjointBase):

    def __init__(self, **kwargs):
        super(CastDisjointToRaggedAttributes, self).__init__(**kwargs)

    def build(self, input_shape):
        self.built = True

    def call(self, inputs, **kwargs):
        r"""Changes node or edge tensors into a Pytorch Geometric (PyG) compatible tensor format.

        Args:
            list: `[attr, graph_id, item_id, item_counts]` .

                - attr (Tensor): Represents attributes or coordinates of shape `([N], F, ...)`
                - graph_id (Tensor): ID tensor of batch assignment in disjoint graph of shape `([N], )` .
                - item_id (Tensor): The ID-tensor to assign each node to its respective graph of shape `([N], )` .
                - item_counts (Tensor): Tensor of lengths for each graph of shape `(batch, )` .

        Returns:
            Tensor: Ragged or Jagged tensor of attributes.
        """
        raise NotImplementedError()


CastDisjointToRaggedAttributes.__init__.__doc__ = CastDisjointToRaggedAttributes.__init__.__doc__


from keras.layers import Dense, Concatenate, Activation, Layer
from kgcnn.layers.gather import GatherNodesIngoing, GatherNodesOutgoing
from kgcnn.layers.aggr import AggregateLocalEdgesAttention


class AttentiveHeadFP_(Layer):
    r"""Computes the attention head for `Attentive FP <https://doi.org/10.1021/acs.jmedchem.9b00959>`__ model.
    The attention coefficients are computed by :math:`a_{ij} = \sigma_1( W_1 [h_i || h_j] )`.
    The initial representation :math:`h_i` and :math:`h_j` must be calculated beforehand.
    The attention is obtained by :math:`\alpha_{ij} = \text{softmax}_j (a_{ij})`.
    And finally pooled for context :math:`C_i = \sigma_2(\sum_j \alpha_{ij} W_2 h_j)`.

    An edge is defined by index tuple :math:`(i, j)` with the direction of the connection from :math:`j` to :math:`i`.
    """

    def __init__(self,
                 units,
                 use_edge_features=False,
                 activation='kgcnn>leaky_relu2',
                 activation_context="elu",
                 use_bias=True,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 **kwargs):
        """Initialize layer.

        Args:
            units (int): Units for the linear trafo of node features before attention.
            use_edge_features (bool): Append edge features to attention computation. Default is False.
            activation (str): Activation. Default is "kgcnn>leaky_relu2".
            activation_context (str): Activation function for context. Default is "elu".
            use_bias (bool): Use bias. Default is True.
            kernel_regularizer: Kernel regularization. Default is None.
            bias_regularizer: Bias regularization. Default is None.
            activity_regularizer: Activity regularization. Default is None.
            kernel_constraint: Kernel constrains. Default is None.
            bias_constraint: Bias constrains. Default is None.
            kernel_initializer: Initializer for kernels. Default is 'glorot_uniform'.
            bias_initializer: Initializer for bias. Default is 'zeros'.
        """
        super(AttentiveHeadFP_, self).__init__(**kwargs)
        # Changes in keras serialization behaviour for activations in 3.0.2.
        # Keep string at least for default. Also renames to prevent clashes with keras leaky_relu.
        if activation in ["kgcnn>leaky_relu", "kgcnn>leaky_relu2"]:
            activation = {"class_name": "function", "config": "kgcnn>leaky_relu2"}
        self.use_edge_features = use_edge_features
        self.units = int(units)
        self.use_bias = use_bias
        kernel_args = {"kernel_regularizer": kernel_regularizer,
                       "activity_regularizer": activity_regularizer, "bias_regularizer": bias_regularizer,
                       "kernel_constraint": kernel_constraint, "bias_constraint": bias_constraint,
                       "kernel_initializer": kernel_initializer, "bias_initializer": bias_initializer}

        self.lay_linear_trafo = Dense(units, activation="linear", use_bias=use_bias, **kernel_args)
        self.lay_alpha_activation = Dense(units, activation=activation, use_bias=use_bias, **kernel_args)
        self.lay_alpha = Dense(1, activation="linear", use_bias=False, **kernel_args)
        self.lay_gather_in = GatherNodesIngoing()
        self.lay_gather_out = GatherNodesOutgoing()
        self.lay_concat = Concatenate(axis=-1)

        self.lay_pool_attention = AggregateLocalEdgesAttention()
        self.lay_final_activ = Activation(activation=activation_context)
        if use_edge_features:
            self.lay_fc1 = Dense(units, activation=activation, use_bias=use_bias, **kernel_args)
            self.lay_fc2 = Dense(units, activation=activation, use_bias=use_bias, **kernel_args)
            self.lay_concat_edge = Concatenate(axis=-1)

    def build(self, input_shape):
        """Build layer."""
        super(AttentiveHeadFP_, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): of [node, edges, edge_indices]

                - nodes (Tensor): Node embeddings of shape ([N], F)
                - edges (Tensor): Edge or message embeddings of shape ([M], F)
                - edge_indices (Tensor): Edge indices referring to nodes of shape (2, [M])

        Returns:
            Tensor: Hidden tensor of pooled edge attentions for each node.
        """
        node, edge, edge_index = inputs

        if self.use_edge_features:
            n_in = self.lay_gather_in([node, edge_index], **kwargs)
            n_out = self.lay_gather_out([node, edge_index], **kwargs)
            n_in = self.lay_fc1(n_in, **kwargs)
            n_out = self.lay_concat_edge([n_out, edge], **kwargs)
            n_out = self.lay_fc2(n_out, **kwargs)
        else:
            n_in = self.lay_gather_in([node, edge_index], **kwargs)
            n_out = self.lay_gather_out([node, edge_index], **kwargs)

        wn_out = self.lay_linear_trafo(n_out, **kwargs)
        e_ij = self.lay_concat([n_in, n_out], **kwargs)
        e_ij = self.lay_alpha_activation(e_ij, **kwargs)  # Maybe uses GAT original definition.
        # a_ij = e_ij
        a_ij = self.lay_alpha(e_ij, **kwargs)  # Should be dimension (batch,None,1) not fully clear in original paper.
        n_i = self.lay_pool_attention([node, wn_out, a_ij, edge_index], **kwargs)
        out = self.lay_final_activ(n_i, **kwargs)
        return out

    def get_config(self):
        """Update layer config."""
        config = super(AttentiveHeadFP_, self).get_config()
        config.update({"use_edge_features": self.use_edge_features, "use_bias": self.use_bias,
                       "units": self.units})
        conf_sub = self.lay_alpha_activation.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation"]:
            if x in conf_sub.keys():
                config.update({x: conf_sub[x]})
        conf_context = self.lay_final_activ.get_config()
        config.update({"activation_context": conf_context["activation"]})
        return config


import keras as ks
from keras.layers import Layer, Dense, Concatenate, GRUCell, Activation
from kgcnn.layers.gather import GatherState
from keras import ops
import kgcnn.ops.activ
from kgcnn.ops.scatter import scatter_reduce_softmax
from kgcnn.layers.aggr import Aggregate


class PoolingNodes(Layer):
    r"""Main layer to pool node or edge attributes. Uses :obj:`Aggregate` layer."""

    def __init__(self, pooling_method="scatter_sum", **kwargs):
        """Initialize layer.

        Args:
            pooling_method (str): Pooling method to use i.e. segment_function. Default is 'scatter_sum'.
        """
        super(PoolingNodes, self).__init__(**kwargs)
        self.pooling_method = pooling_method
        self._to_aggregate = Aggregate(pooling_method=pooling_method)

    def build(self, input_shape):
        """Build Layer."""
        self._to_aggregate.build([input_shape[1], input_shape[2], input_shape[0]])
        self.built = True

    def compute_output_shape(self, input_shape):
        """Compute output shape."""
        return self._to_aggregate.compute_output_shape([input_shape[1], input_shape[2], input_shape[0]])

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [reference, attr, weights, batch_index]

                - reference (Tensor): Reference for aggregation of shape `(batch, ...)` .
                - attr (Tensor): Node or edge embeddings of shape `([N], F)` .
                - batch_index (Tensor): Batch assignment of shape `([N], )` .

        Returns:
            Tensor: Embedding tensor of pooled node of shape `(batch, F)` .
        """
        reference, x, idx = inputs
        return self._to_aggregate([x, idx, reference])

    def get_config(self):
        """Update layer config."""
        config = super(PoolingNodes, self).get_config()
        config.update({"pooling_method": self.pooling_method})
        return config


class PoolingWeightedNodes(Layer):
    r"""Weighted polling all embeddings of edges or nodes per batch to obtain a graph level embedding.

    .. note::

        In addition to pooling embeddings a weight tensor must be supplied that scales each embedding before
        pooling. Must broadcast.
    """

    def __init__(self, pooling_method="scatter_sum", **kwargs):
        """Initialize layer.

        Args:
            pooling_method (str): Pooling method to use i.e. segment_function. Default is 'scatter_sum'.
        """
        super(PoolingWeightedNodes, self).__init__(**kwargs)
        self.pooling_method = pooling_method
        self._to_aggregate = Aggregate(pooling_method=pooling_method)

    def build(self, input_shape):
        """Build layer."""
        assert len(input_shape) == 4
        ref_shape, attr_shape, weights_shape, index_shape = [list(x) for x in input_shape]
        self._to_aggregate.build([tuple(x) for x in [attr_shape, index_shape, ref_shape]])
        self.built = True

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [reference, attr, weights, batch_index]

                - reference (Tensor): Reference for aggregation of shape `(batch, ...)` .
                - attr (Tensor): Node or edge embeddings of shape `([N], F)` .
                - weights (Tensor): Node or message weights. Most broadcast to nodes. Shape ([N], 1).
                - batch_index (Tensor): Batch assignment of shape `([N], )` .

        Returns:
            Tensor: Embedding tensor of pooled node of shape `(batch, F)` .
        """
        # Need ragged input but can be generalized in the future.
        reference, x, w, idx = inputs
        xw = ops.broadcast_to(ops.cast(w, dtype=x.dtype), ops.shape(x)) * x
        return self._to_aggregate([xw, idx, reference])

    def get_config(self):
        """Update layer config."""
        config = super(PoolingWeightedNodes, self).get_config()
        config.update({"pooling_method": self.pooling_method})
        return config


class PoolingEmbeddingAttention(Layer):
    r"""Polling all embeddings of edges or nodes per batch to obtain a graph level embedding in form of a
    :obj:`Tensor` .

    Uses attention for pooling. i.e. :math:`s =  \sum_j \alpha_{i} n_i` .
    The attention is computed via: :math:`\alpha_i = \text{softmax}_i(a_i)` from the attention
    coefficients :math:`a_i` .
    The attention coefficients must be computed beforehand by node or edge features or by :math:`\sigma( W [s || n_i])`
    and are passed to this layer as input. Thereby this layer has no weights and only does pooling.
    In summary, :math:`s =  \sum_i \text{softmax}_j(a_i) n_i` is computed by the layer.
    """

    def __init__(self,
                 softmax_method="scatter_softmax",
                 pooling_method="scatter_sum",
                 normalize_softmax: bool = False,
                 **kwargs):
        """Initialize layer.

        Args:
            normalize_softmax (bool): Whether to use normalize in softmax. Default is False.
        """
        super(PoolingEmbeddingAttention, self).__init__(**kwargs)
        self.normalize_softmax = normalize_softmax
        self.pooling_method = pooling_method
        self.softmax_method = softmax_method
        self.to_aggregate = Aggregate(pooling_method=pooling_method)

    def build(self, input_shape):
        """Build layer."""
        assert len(input_shape) == 4
        ref_shape, attr_shape, attention_shape, index_shape = [list(x) for x in input_shape]
        self.to_aggregate.build([tuple(x) for x in [attr_shape, index_shape, ref_shape]])
        self.built = True

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [reference, attr, attention, batch_index]

                - reference (Tensor): Reference for aggregation of shape `(batch, ...)` .
                - attr (Tensor): Node or edge embeddings of shape `([N], F)` .
                - attention (Tensor): Attention coefficients of shape `([N], 1)` .
                - batch_index (Tensor): Batch assignment of shape `([N], )` .

        Returns:
            Tensor: Embedding tensor of pooled node of shape `(batch, F)` .
        """
        reference, attr, attention, batch_index = inputs
        shape_attention = ops.shape(reference)[:1] + ops.shape(attention)[1:]
        a = scatter_reduce_softmax(batch_index, attention, shape=shape_attention, normalize=self.normalize_softmax)
        x = attr * ops.broadcast_to(a, ops.shape(attr))
        return self.to_aggregate([x, batch_index, reference])

    def get_config(self):
        """Update layer config."""
        config = super(PoolingEmbeddingAttention, self).get_config()
        config.update({
            "normalize_softmax": self.normalize_softmax, "pooling_method": self.pooling_method,
            "softmax_method": self.softmax_method
        })
        return config


PoolingNodesAttention = PoolingEmbeddingAttention


class PoolingNodesAttentive(Layer):
    r"""Computes the attentive pooling for node embeddings for
    `Attentive FP <https://doi.org/10.1021/acs.jmedchem.9b00959>`__ model.
    """

    def __init__(self,
                 units,
                 depth=3,
                 pooling_method="sum",
                 activation="kgcnn>leaky_relu2",
                 activation_context="elu",
                 use_bias=True,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 recurrent_activation='sigmoid',
                 recurrent_initializer='orthogonal',
                 recurrent_regularizer=None,
                 recurrent_constraint=None,
                 dropout=0.0,
                 recurrent_dropout=0.0,
                 reset_after=True,
                 **kwargs):
        """Initialize layer.
        
        Args:
            units (int): Units for the linear trafo of node features before attention.
            pooling_method(str): Initial pooling before iteration. Default is "sum".
            depth (int): Number of iterations for graph embedding. Default is 3.
            activation (str): Activation. Default is "kgcnn>leaky_relu2".
            activation_context (str): Activation function for context. Default is "elu".
            use_bias (bool): Use bias. Default is True.
            kernel_regularizer: Kernel regularization. Default is None.
            bias_regularizer: Bias regularization. Default is None.
            activity_regularizer: Activity regularization. Default is None.
            kernel_constraint: Kernel constrains. Default is None.
            bias_constraint: Bias constrains. Default is None.
            kernel_initializer: Initializer for kernels. Default is 'glorot_uniform'.
            bias_initializer: Initializer for bias. Default is 'zeros'.
        """
        super(PoolingNodesAttentive, self).__init__(**kwargs)
        # Changes in keras serialization behaviour for activations in 3.0.2.
        # Keep string at least for default. Also renames to prevent clashes with keras leaky_relu.
        if activation in ["kgcnn>leaky_relu", "kgcnn>leaky_relu2"]:
            activation = {"class_name": "function", "config": "kgcnn>leaky_relu2"}
        self.pooling_method = pooling_method
        self.depth = depth
        self.units = int(units)
        kernel_args = {"use_bias": use_bias, "kernel_regularizer": kernel_regularizer,
                       "activity_regularizer": activity_regularizer, "bias_regularizer": bias_regularizer,
                       "kernel_constraint": kernel_constraint, "bias_constraint": bias_constraint,
                       "kernel_initializer": kernel_initializer, "bias_initializer": bias_initializer}
        gru_args = {"recurrent_activation": recurrent_activation,
                    "use_bias": use_bias, "kernel_initializer": kernel_initializer,
                    "recurrent_initializer": recurrent_initializer, "bias_initializer": bias_initializer,
                    "kernel_regularizer": kernel_regularizer, "recurrent_regularizer": recurrent_regularizer,
                    "bias_regularizer": bias_regularizer, "kernel_constraint": kernel_constraint,
                    "recurrent_constraint": recurrent_constraint, "bias_constraint": bias_constraint,
                    "dropout": dropout, "recurrent_dropout": recurrent_dropout, "reset_after": reset_after}

        self.lay_linear_trafo = Dense(units, activation="linear", **kernel_args)
        self.lay_alpha = Dense(1, activation=activation, **kernel_args)
        self.lay_gather_s = GatherState()
        self.lay_concat = Concatenate(axis=-1)
        self.lay_pool_start = PoolingNodes(pooling_method=self.pooling_method)
        self.lay_pool_attention = PoolingNodesAttention()
        self.lay_final_activ = Activation(activation=activation_context)
        self.lay_gru = GRUCell(units=units, activation="tanh", **gru_args)

    def build(self, input_shape):
        """Build layer."""
        super(PoolingNodesAttentive, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs: [reference, nodes, batch_index]

                - reference (Tensor): Reference for aggregation of shape `(batch, ...)` .
                - nodes (Tensor): Node embeddings of shape `([N], F)` .
                - batch_index (Tensor): Batch assignment of shape `([N], )` .

        Returns:
            Tensor: Hidden tensor of pooled node attentions of shape (batch, F).
        """
        ref, node, batch_index = inputs

        h = self.lay_pool_start([ref, node, batch_index], **kwargs)
        wn = self.lay_linear_trafo(node, **kwargs)
        for _ in range(self.depth):
            hv = self.lay_gather_s([h, batch_index], **kwargs)
            ev = self.lay_concat([hv, node], **kwargs)
            av = self.lay_alpha(ev, **kwargs)
            cont = self.lay_pool_attention([ref, wn, av, batch_index], **kwargs)
            cont = self.lay_final_activ(cont, **kwargs)
            h, _ = self.lay_gru(cont, h, **kwargs)

        out = h
        return out

    def get_config(self):
        """Update layer config."""
        config = super(PoolingNodesAttentive, self).get_config()
        config.update({"units": self.units, "depth": self.depth, "pooling_method": self.pooling_method})
        conf_sub = self.lay_alpha.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation", "use_bias"]:
            if x in conf_sub.keys():
                config.update({x: conf_sub[x]})
        conf_context = self.lay_final_activ.get_config()
        config.update({"activation_context": conf_context["activation"]})
        conf_gru = self.lay_gru.get_config()
        for x in ["recurrent_activation", "recurrent_initializer", "recurrent_regularizer", "recurrent_constraint",
                  "dropout", "recurrent_dropout", "reset_after"]:
            if x in conf_gru.keys():
                config.update({x: conf_gru[x]})
        return config



import tree
from keras import KerasTensor
from keras.backend import backend


def any_symbolic_tensors(args=None, kwargs=None):
    args = args or ()
    kwargs = kwargs or {}
    for x in tree.flatten((args, kwargs)):
        if isinstance(x, KerasTensor):
            return True
    return False


# Import backend functions.
if backend() == "tensorflow":
    from kgcnn.backend._tensorflow import *
elif backend() == "jax":
    from kgcnn.backend._jax import *
elif backend() == "torch":
    from kgcnn.backend._torch import *
elif backend() == "numpy":
    from kgcnn.backend._numpy import *
else:
    raise ValueError(f"Unable to import backend : {backend()}")



import keras as ks
# import keras_core.saving
from keras.layers import Layer
from keras import ops
from kgcnn.ops.scatter import (
    scatter_reduce_min, scatter_reduce_mean, scatter_reduce_max, scatter_reduce_sum, scatter_reduce_softmax)
from kgcnn import __indices_axis__ as global_axis_indices
from kgcnn import __index_receive__ as global_index_receive


@ks.saving.register_keras_serializable(package='kgcnn', name='Aggregate')
class Aggregate(Layer):  # noqa
    """Main class for aggregating node or edge features.

    The class essentially uses a reduce function by name to aggregate a feature list given indices to group by.
    Possible supported permutation invariant aggregations are 'sum', 'mean', 'max' or 'min'.
    For aggregation either scatter or segment operation can be used from the backend, if available.
    Note that you have to specify which to use with e.g. 'scatter_sum'.
    This layer further requires a reference tensor to either statically infer the output shape or even directly
    aggregate the values into.
    """

    def __init__(self, pooling_method: str = "scatter_sum", axis=0, **kwargs):
        """Initialize layer.

        Args:
            pooling_method (str): Method for aggregation. Default is 'scatter_sum'.
            axis (int): Axis to aggregate. Default is 0.
        """
        super(Aggregate, self).__init__(**kwargs)
        # Shorthand notation
        if pooling_method == "sum":
            pooling_method = "scatter_sum"
        if pooling_method == "max":
            pooling_method = "scatter_max"
        if pooling_method == "min":
            pooling_method = "scatter_min"
        if pooling_method == "mean":
            pooling_method = "scatter_mean"
        self.pooling_method = pooling_method
        self.axis = axis
        if axis != 0:
            raise NotImplementedError("Only aggregate at axis=0 is supported for `Aggregate` layer.")
        pooling_by_name = {
            "scatter_sum": scatter_reduce_sum,
            "scatter_mean": scatter_reduce_mean,
            "scatter_max": scatter_reduce_max,
            "scatter_min": scatter_reduce_min,
            "segment_sum": None,
            "segment_mean": None,
            "segment_max": None,
            "segment_min": None
        }
        self._pool_method = pooling_by_name[pooling_method]
        self._use_scatter = "scatter" in pooling_method
        self._use_reference_for_aggregation = "update" in pooling_method

    def build(self, input_shape):
        """Build layer."""
        # Nothing to build here. No sub-layers.
        self.built = True

    def compute_output_shape(self, input_shape):
        """Compute output shape."""
        assert len(input_shape) == 3
        x_shape, _, dim_size = input_shape
        return tuple(list(dim_size[:1]) + list(x_shape[1:]))

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [values, indices, reference]

                - values (Tensor): Values to aggregate of shape `(M, ...)`.
                - indices (Tensor): Indices of target assignment of shape `(M, )`.
                - reference (Tensor): Target reference tensor of shape `(N, ...)`.

        Returns:
            Tensor: Aggregated values of shape `(N, ...)`.
        """
        x, index, reference = inputs
        shape = ops.shape(reference)[:1] + ops.shape(x)[1:]
        if self._use_scatter:
            return self._pool_method(index, x, shape=shape)
        else:
            raise NotImplementedError()

    def get_config(self):
        """Get config for layer."""
        conf = super(Aggregate, self).get_config()
        conf.update({"pooling_method": self.pooling_method, "axis": self.axis})
        return conf


@ks.saving.register_keras_serializable(package='kgcnn', name='AggregateLocalEdges')
class AggregateLocalEdges(Layer):
    r"""The main aggregation or pooling layer to collect all edges or edge-like embeddings per node,
    corresponding to the receiving node, which is defined by edge indices.

    Apply e.g. 'sum' or 'mean' on edges with same target ID taken from the (edge) index tensor, that has a list of
    all connections as :math:`(i, j)` .

    In the default definition for this layer index :math:`i` is expected to be the
    receiving or target node (in standard case of directed edges). This can be changed by setting :obj:`pooling_index` ,
    i.e. `index_tensor[pooling_index]` to get the indices to aggregate the edges with.
    This layers uses the :obj:`Aggregate` layer and its functionality.
    """
    def __init__(self,
                 pooling_method="scatter_sum",
                 pooling_index: int = global_index_receive,
                 axis_indices: int = global_axis_indices,
                 **kwargs):
        """Initialize layer.

        Args:
            pooling_method (str): Pooling method to use i.e. segment_function. Default is 'scatter_sum'.
            pooling_index (int): Index to pick IDs for pooling edge-like embeddings. Default is 0.
            axis_indices (bool): The axis of the index tensor to pick IDs from. Default is 0.
        """
        super(AggregateLocalEdges, self).__init__(**kwargs)
        self.pooling_index = pooling_index
        self.pooling_method = pooling_method
        self.to_aggregate = Aggregate(pooling_method=pooling_method)
        self.axis_indices = axis_indices

    def build(self, input_shape):
        """Build layer."""
        # Layer has no variables but still can call build on sub-layers.
        assert len(input_shape) == 3
        node_shape, edges_shape, edge_index_shape = [list(x) for x in input_shape]
        edge_index_shape.pop(self.axis_indices)
        self.to_aggregate.build([tuple(x) for x in [edges_shape, edge_index_shape, node_shape]])
        self.built = True

    def compute_output_shape(self, input_shape):
        """Compute output shape."""
        assert len(input_shape) == 3
        node_shape, edges_shape, edge_index_shape = [list(x) for x in input_shape]
        edge_index_shape.pop(self.axis_indices)
        return self.to_aggregate.compute_output_shape([tuple(x) for x in [edges_shape, edge_index_shape, node_shape]])

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs (list): [reference, values, indices]

                - reference (Tensor): Target reference tensor of shape `(N, ...)`.
                - values (Tensor): Values to aggregate of shape `(M, ...)`.
                - indices (Tensor): Indices of edges of shape `(2, M, )`.

        Returns:
            Tensor: Aggregated values of shape `(N, ...)`.
        """
        n, edges, edge_index = inputs
        receive_indices = ops.take(edge_index, self.pooling_index, axis=self.axis_indices)
        return self.to_aggregate([edges, receive_indices, n])

    def get_config(self):
        """Update layer config."""
        conf = super(AggregateLocalEdges, self).get_config()
        conf.update({"pooling_index": self.pooling_index, "pooling_method": self.pooling_method,
                     "axis_indices": self.axis_indices})
        return conf


@ks.saving.register_keras_serializable(package='kgcnn', name='AggregateWeightedLocalEdges')
class AggregateWeightedLocalEdges(AggregateLocalEdges):
    r"""This class inherits from :obj:`AggregateLocalEdges` for aggregating weighted edges.

    Please check the documentation of :obj:`AggregateLocalEdges` for more information.

    .. note::

        In addition to aggregating edge embeddings a weight tensor must be supplied that scales each edge before
        pooling. Must broadcast.
    """

    def __init__(self, pooling_method: str = "scatter_sum", pooling_index: int = global_index_receive,
                 normalize_by_weights: bool = False, axis_indices: int = global_axis_indices, **kwargs):
        """Initialize layer.

        Args:
            pooling_method (str): Pooling method to use i.e. segment_function. Default is 'scatter_sum'.
            normalize_by_weights (bool): Whether to normalize pooled features by the sum of weights. Default is False.
            pooling_index (int): Index to pick IDs for pooling edge-like embeddings. Default is 0.
            axis_indices (bool): The axis of the index tensor to pick IDs from. Default is 0.
        """
        super(AggregateWeightedLocalEdges, self).__init__(**kwargs)
        self.normalize_by_weights = normalize_by_weights
        self.pooling_index = pooling_index
        self.pooling_method = pooling_method
        # to_aggregate already made by super
        if self.normalize_by_weights:
            self.to_aggregate_weights = Aggregate(pooling_method="scatter_sum")
        self.axis_indices = axis_indices

    def build(self, input_shape):
        """Build layer."""
        assert len(input_shape) == 4
        node_shape, edges_shape, edge_index_shape, weights_shape = [list(x) for x in input_shape]
        edge_index_shape.pop(self.axis_indices)
        self.to_aggregate.build([tuple(x) for x in [edges_shape, edge_index_shape, node_shape]])
        if self.normalize_by_weights:
            self.to_aggregate_weights.build([tuple(x) for x in [weights_shape, edge_index_shape, node_shape]])
        self.built = True

    def compute_output_shape(self, input_shape):
        """Compute output shape."""
        assert len(input_shape) == 4
        node_shape, edges_shape, edge_index_shape, weights_shape = [list(x) for x in input_shape]
        edge_index_shape.pop(self.axis_indices)
        return self.to_aggregate.compute_output_shape([tuple(x) for x in [edges_shape, edge_index_shape, node_shape]])

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs (list): [reference, values, indices, weights]

                - reference (Tensor): Target reference tensor of shape `(N, ...)`.
                - values (Tensor): Values to aggregate of shape `(M, ...)`.
                - indices (Tensor): Indices of edges of shape `(2, M, )`.
                - weights (Tensor): Weight tensor for values of shape `(M, ...)`.

        Returns:
            Tensor: Aggregated values of shape `(N, ...)`.
        """
        n, edges, edge_index, weights = inputs
        edges = edges*weights
        receive_indices = ops.take(edge_index, self.pooling_index, axis=self.axis_indices)
        out = self.to_aggregate([edges, receive_indices, n])

        if self.normalize_by_weights:
            norm = self.to_aggregate_weights([weights, receive_indices, n])
            out = out/norm
        return out

    def get_config(self):
        """Update layer config."""
        conf = super(AggregateWeightedLocalEdges, self).get_config()
        conf.update({"pooling_index": self.pooling_index, "pooling_method": self.pooling_method,
                     "axis_indices": self.axis_indices, "normalize_by_weights": self.normalize_by_weights})
        return conf


@ks.saving.register_keras_serializable(package='kgcnn', name='AggregateLocalEdgesAttention')
class AggregateLocalEdgesAttention(Layer):
    r"""Aggregate local edges via Attention mechanism.
    Uses attention for pooling. i.e. :math:`n_i =  \sum_j \alpha_{ij} e_{ij}`
    The attention is computed via: :math:`\alpha_ij = \text{softmax}_j (a_{ij})` from the
    attention coefficients :math:`a_{ij}` .

    The attention coefficients must be computed beforehand by edge features or by :math:`\sigma( W n_i || W n_j)` and
    are passed to this layer as input. Thereby this layer has no weights and only does pooling.
    In summary, the following is computed by the layer:

    .. math::

            n_i = \sum_j \text{softmax}_j (a_{ij}) e_{ij}
    """

    def __init__(self,
                 softmax_method="scatter_softmax",
                 pooling_method="scatter_sum",
                 pooling_index: int = global_index_receive,
                 is_sorted: bool = False,
                 has_unconnected: bool = True,
                 normalize_softmax: bool = False,
                 axis_indices: int = global_axis_indices,
                 **kwargs):
        """Initialize layer.

        Args:
            softmax_method (str): Method to apply softmax to attention coefficients. Default is 'scatter_softmax'.
            pooling_method (str): Pooling method for this layer. Default is 'scatter_sum'.
            pooling_index (int): Index to pick ID's for pooling edge-like embeddings. Default is 0.
            is_sorted (bool): If the edge indices are sorted for first ingoing index. Default is False.
            has_unconnected (bool): If unconnected nodes are allowed. Default is True.
            normalize_softmax (bool): Whether to use normalize in softmax. Default is False.
            axis_indices (int): The axis of the index tensor to pick IDs from. Default is 0.
        """
        super(AggregateLocalEdgesAttention, self).__init__(**kwargs)
        self.pooling_method = pooling_method
        self.pooling_index = pooling_index
        self.is_sorted = is_sorted
        self.has_unconnected = has_unconnected
        self.normalize_softmax = normalize_softmax
        self.softmax_method = softmax_method
        self.to_aggregate = Aggregate(pooling_method=pooling_method)
        self.axis_indices = axis_indices

    def build(self, input_shape):
        """Build layer."""
        assert len(input_shape) == 4
        node_shape, edges_shape, attention_shape, edge_index_shape = [list(x) for x in input_shape]
        edge_index_shape.pop(self.axis_indices)
        self.to_aggregate.build([tuple(x) for x in [edges_shape, edge_index_shape, node_shape]])
        self.built = True

    def compute_output_shape(self, input_shape):
        """Compute output shape."""
        assert len(input_shape) == 4
        node_shape, edges_shape, attention_shape, edge_index_shape = [list(x) for x in input_shape]
        edge_index_shape.pop(self.axis_indices)
        return self.to_aggregate.compute_output_shape([tuple(x) for x in [edges_shape, edge_index_shape, node_shape]])

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [node, edges, attention, edge_indices]

                - nodes (Tensor): Node embeddings of shape `(N, F)`
                - edges (Tensor): Edge or message embeddings of shape `(M, F)`
                - attention (Tensor): Attention coefficients of shape `(M, 1)`
                - edge_indices (Tensor): Edge indices referring to nodes of shape `(2, M)`

        Returns:
            Tensor: Embedding tensor of aggregated edge attentions for each node of shape `(N, F)` .
        """
        reference, x, attention, edge_index = inputs
        receive_indices = ops.take(edge_index, self.pooling_index, axis=self.axis_indices)
        shape_attention = ops.shape(reference)[:1] + ops.shape(attention)[1:]
        a = scatter_reduce_softmax(receive_indices, attention, shape=shape_attention, normalize=self.normalize_softmax)
        x = x * ops.broadcast_to(a, ops.shape(x))
        return self.to_aggregate([x, receive_indices, reference])

    def get_config(self):
        """Update layer config."""
        config = super(AggregateLocalEdgesAttention, self).get_config()
        config.update({
            "normalize_softmax": self.normalize_softmax, "pooling_method": self.pooling_method,
            "pooling_index": self.pooling_index, "axis_indices": self.axis_indices,
            "softmax_method": self.softmax_method, "is_sorted": self.is_sorted, "has_unconnected": self.has_unconnected
        })
        return config


@ks.saving.register_keras_serializable(package='kgcnn', name='AggregateLocalEdgesLSTM')
class AggregateLocalEdgesLSTM(Layer):
    r"""Aggregating edges via a LSTM.

    Apply LSTM on edges with same target ID taken from the (edge) index tensor. Uses keras LSTM layer internally.

    .. note::

        Must provide a max length of edges per nodes, since keras LSTM requires padded input. Also required for use
        in connection with jax backend.
    """

    def __init__(self,
                 units: int,
                 max_edges_per_node: int,
                 pooling_method="LSTM",
                 pooling_index=global_index_receive,
                 axis_indices: int = global_axis_indices,
                 activation='tanh', recurrent_activation='sigmoid',
                 use_bias=True, kernel_initializer='glorot_uniform',
                 recurrent_initializer='orthogonal',
                 bias_initializer='zeros', unit_forget_bias=True,
                 kernel_regularizer=None, recurrent_regularizer=None, bias_regularizer=None,
                 activity_regularizer=None, kernel_constraint=None, recurrent_constraint=None,
                 bias_constraint=None, dropout=0.0, recurrent_dropout=0.0,
                 return_sequences=False, return_state=False, go_backwards=False, stateful=False,
                 time_major=False, unroll=False,
                 **kwargs):
        """Initialize layer.

        Args:
            units (int): Units for LSTM cell.
            max_edges_per_node (int): Max number of edges per node.
            pooling_method (str): Pooling method. Default is 'LSTM'.
            pooling_index (int): Index to pick IDs for pooling edge-like embeddings. Default is 0.
            axis_indices (int): Axis to pick receiving index from. Default is 0.
            activation: Activation function to use.
                Default: hyperbolic tangent (`tanh`). If you pass `None`, no activation
                is applied (ie. "linear" activation: `a(x) = x`).
            recurrent_activation: Activation function to use for the recurrent step.
                Default: sigmoid (`sigmoid`). If you pass `None`, no activation is
                applied (ie. "linear" activation: `a(x) = x`).
            use_bias: Boolean (default `True`), whether the layer uses a bias vector.
            kernel_initializer: Initializer for the `kernel` weights matrix, used for
                the linear transformation of the inputs. Default: `glorot_uniform`.
                recurrent_initializer: Initializer for the `recurrent_kernel` weights
                matrix, used for the linear transformation of the recurrent state.
                Default: `orthogonal`.
            bias_initializer: Initializer for the bias vector. Default: `zeros`.
                unit_forget_bias: Boolean (default `True`). If True, add 1 to the bias of
                the forget gate at initialization. Setting it to true will also force
                `bias_initializer="zeros"`. This is recommended in [Jozefowicz et
                al.](http://www.jmlr.org/proceedings/papers/v37/jozefowicz15.pdf).
            kernel_regularizer: Regularizer function applied to the `kernel` weights
                matrix. Default: `None`.
            recurrent_regularizer: Regularizer function applied to the
                `recurrent_kernel` weights matrix. Default: `None`.
                bias_regularizer: Regularizer function applied to the bias vector. Default:
                `None`.
            activity_regularizer: Regularizer function applied to the output of the
                layer (its "activation"). Default: `None`.
            kernel_constraint: Constraint function applied to the `kernel` weights
                matrix. Default: `None`.
            recurrent_constraint: Constraint function applied to the `recurrent_kernel`
                weights matrix. Default: `None`.
            bias_constraint: Constraint function applied to the bias vector. Default:
                `None`.
            dropout: Float between 0 and 1. Fraction of the units to drop for the linear
                transformation of the inputs. Default: 0.
                recurrent_dropout: Float between 0 and 1. Fraction of the units to drop for
                the linear transformation of the recurrent state. Default: 0.
            return_sequences: Boolean. Whether to return the last output. in the output
                sequence, or the full sequence. Default: `False`.
            return_state: Boolean. Whether to return the last state in addition to the
                output. Default: `False`.
            go_backwards: Boolean (default `False`). If True, process the input sequence
                backwards and return the reversed sequence.
            stateful: Boolean (default `False`). If True, the last state for each sample
                at index i in a batch will be used as initial state for the sample of
                index i in the following batch.
            time_major: The shape format of the `inputs` and `outputs` tensors.
                If True, the inputs and outputs will be in shape
                `[timesteps, batch, feature]`, whereas in the False case, it will be
                `[batch, timesteps, feature]`. Using `time_major = True` is a bit more
                efficient because it avoids transposes at the beginning and end of the
                RNN calculation. However, most TensorFlow data is batch-major, so by
                default this function accepts input and emits output in batch-major
                form.
            unroll: Boolean (default `False`). If True, the network will be unrolled,
                else a symbolic loop will be used. Unrolling can speed-up a RNN, although
                it tends to be more memory-intensive. Unrolling is only suitable for short
                sequences.
        """
        super(AggregateLocalEdgesLSTM, self).__init__(**kwargs)
        self.pooling_method = pooling_method
        self.axis_indices = axis_indices
        self.pooling_index = pooling_index
        self.lstm_unit = ks.layers.LSTM(
            units=units, activation=activation, recurrent_activation=recurrent_activation, use_bias=use_bias,
            kernel_initializer=kernel_initializer, recurrent_initializer=recurrent_initializer,
            bias_initializer=bias_initializer, unit_forget_bias=unit_forget_bias, kernel_regularizer=kernel_regularizer,
            recurrent_regularizer=recurrent_regularizer, bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer, kernel_constraint=kernel_constraint,
            recurrent_constraint=recurrent_constraint, bias_constraint=bias_constraint, dropout=dropout,
            recurrent_dropout=recurrent_dropout, return_sequences=return_sequences, return_state=return_state,
            go_backwards=go_backwards, stateful=stateful, time_major=time_major, unroll=unroll
        )
        if self.pooling_method not in ["LSTM", "lstm"]:
            raise ValueError(
                "Aggregate method does not match layer, expected 'LSTM' but got '%s'." % self.pooling_method)
        self.max_edges_per_node = max_edges_per_node

    def build(self, input_shape):
        """Build layer."""
        super(AggregateLocalEdgesLSTM, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [node, edges, edge_indices, graph_id_edge]

                - nodes (Tensor): Node embeddings of shape `(N, F)`
                - edges (Tensor): Edge or message embeddings of shape `(M, F)`
                - edge_indices (Tensor): Edge indices referring to nodes of shape `(2, M)`
                - graph_id_edge (Tensor): Graph ID for each edge of shape `(M, )`

        Returns:
            Tensor: Embedding tensor of aggregated edges for each node of shape `(N, F)` .
        """
        n, edges, edge_index, edge_id = inputs
        receive_indices = ops.take(edge_index, self.pooling_index, axis=self.axis_indices)

        dim_n = ops.shape(n)[0]
        dim_e_per_n = self.max_edges_per_node if self.max_edges_per_node is not None else 2*dim_n+1
        indices = receive_indices * ops.convert_to_tensor(dim_e_per_n, dtype=receive_indices.dtype) + ops.cast(
            edge_id, dtype=receive_indices.dtype)
        lstm_input = scatter_reduce_sum(
            indices, edges,
            shape=tuple([dim_n*dim_e_per_n] + list(ops.shape(edges)[1:]))
        )
        lstm_mask = ops.cast(scatter_reduce_sum(
            indices, ops.ones(ops.shape(edges)[:1], dtype=ks.backend.floatx()),
            shape=tuple([dim_n*dim_e_per_n])
        ), dtype="bool")

        lstm_input = ops.reshape(lstm_input, tuple([dim_n, dim_e_per_n] + list(ops.shape(edges)[1:])))
        lstm_mask = ops.reshape(lstm_mask, tuple([dim_n, dim_e_per_n]))

        out = self.lstm_unit(lstm_input, mask=lstm_mask)
        return out

    def get_config(self):
        """Update layer config."""
        config = super(AggregateLocalEdgesLSTM, self).get_config()
        conf_lstm = self.lstm_unit.get_config()
        lstm_param = ["activation", "recurrent_activation", "use_bias", "kernel_initializer", "recurrent_initializer",
                      "bias_initializer", "unit_forget_bias", "kernel_regularizer", "recurrent_regularizer",
                      "bias_regularizer", "activity_regularizer", "kernel_constraint", "recurrent_constraint",
                      "bias_constraint", "dropout", "recurrent_dropout", "implementation", "return_sequences",
                      "return_state", "go_backwards", "stateful", "time_major", "unroll"]
        for x in lstm_param:
            if x in conf_lstm:
                config.update({x: conf_lstm[x]})
        config.update({"pooling_method": self.pooling_method, "axis_indices": self.axis_indices,
                       "pooling_index": self.pooling_index, "max_edges_per_node": self.max_edges_per_node})
        return config


@ks.saving.register_keras_serializable(package='kgcnn', name='RelationalAggregateLocalEdges')
class RelationalAggregateLocalEdges(Layer):
    r"""Layer :obj:`RelationalAggregateLocalEdges` for aggregating relational edges.

    Please check the documentation of :obj:`AggregateLocalEdges` for more information.

    The main aggregation or pooling layer to collect all edges or edge-like embeddings per node, per relation,
    corresponding to the receiving node, which is defined by edge indices.

    .. note::

        An edge relation tensor must be provided which specifies the relation for each edge.
    """

    def __init__(self,
                 num_relations: int,
                 pooling_method="scatter_sum",
                 pooling_index: int = global_index_receive,
                 axis_indices: int = global_axis_indices,
                 **kwargs):
        """Initialize layer.

        Args:
            num_relations (int): Number of possible relations.
            pooling_method (str): Pooling method to use i.e. segment_function. Default is 'mean'.
            pooling_index (int): Index from edge_indices to pick ID's for pooling edge-like embeddings. Default is 0.
            axis_indices (bool): The axis of the index tensor to pick IDs from. Default is 0.
        """
        super(RelationalAggregateLocalEdges, self).__init__(**kwargs)
        self.num_relations = num_relations
        self.pooling_index = pooling_index
        self.pooling_method = pooling_method
        self.axis_indices = axis_indices
        self.to_aggregate = Aggregate(pooling_method=pooling_method)

    def build(self, input_shape):
        """Build layer."""
        super(RelationalAggregateLocalEdges, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs (list): of [node, edges, tensor_index, edge_relation]

                - node (Tensor): Node reference of shape ([N], R, F)
                - edges (Tensor): Edge or message features of shape ([M], F)
                - tensor_index (Tensor): Edge indices referring to nodes of shape (2, [M])
                - edge_relation (Tensor): Edge relation for each edge of shape ([M], )

        Returns:
            Tensor: Aggregated feature tensor of edge features for each node of shape ([N], R, F) .
        """
        nodes, edges, disjoint_indices, edge_relations = inputs
        node_shape = list(ops.shape(nodes))

        receive_indices = ops.take(disjoint_indices, self.pooling_index, axis=self.axis_indices)
        relations = ops.cast(edge_relations, receive_indices.dtype)

        if self.to_aggregate._use_reference_for_aggregation:
            out_tensor = ops.reshape(nodes, [node_shape[0]*node_shape[1]] + node_shape[2:])
            shifts = node_shape[1]
        else:
            out_tensor = ops.repeat(nodes, self.num_relations, axis=0)
            shifts = self.num_relations

        scatter_indices = receive_indices*shifts + relations

        out = self.to_aggregate([edges, scatter_indices, out_tensor])
        out_shape = list(ops.shape(out))

        return ops.reshape(out, [node_shape[0], shifts] + out_shape[1:])

    def get_config(self):
        """Update layer config."""
        config = super(RelationalAggregateLocalEdges, self).get_config()
        config.update({"num_relations": self.num_relations})
        config.update({"pooling_index": self.pooling_index, "pooling_method": self.pooling_method,
                       "axis_indices": self.axis_indices})
        return config



import keras as ks
from kgcnn.layers.gather import GatherNodesOutgoing, GatherEdgesPairs
from kgcnn.layers.aggr import AggregateLocalEdges
from keras.layers import Subtract


class DMPNNPPoolingEdgesDirected(ks.layers.Layer):  # noqa
    """Pooling of edges for around a target node as defined by

    `DMPNN <https://pubs.acs.org/doi/full/10.1021/acs.jcim.9b00237>`__ . This is slightly different as the normal node
    aggregation from message passing like networks. Requires edge pair indices for this implementation.
    """

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(DMPNNPPoolingEdgesDirected, self).__init__(**kwargs)
        self.pool_edge_1 = AggregateLocalEdges(pooling_method="scatter_sum")
        self.gather_edges = GatherNodesOutgoing()
        self.gather_pairs = GatherEdgesPairs()
        self.subtract_layer = Subtract()

    def build(self, input_shape):
        super(DMPNNPPoolingEdgesDirected, self).build(input_shape)
        # Could call build on sub-layers but is not necessary.

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs: [nodes, edges, edge_index, edge_reverse_pair]

                - nodes (Tensor): Node embeddings of shape ([N], F)
                - edges (Tensor): Edge or message embeddings of shape ([M], F)
                - edge_index (Tensor): Edge indices referring to nodes of shape (2, [M])
                - edge_reverse_pair (Tensor): Pair mappings for reverse edges (1, [M])

        Returns:
            Tensor: Edge embeddings of shape ([M], F)
        """
        n, ed, edi, edp = inputs
        pool_edge_receive = self.pool_edge_1([n, ed, edi], **kwargs)  # Sum pooling of all edges
        ed_new = self.gather_edges([pool_edge_receive, edi], **kwargs)
        ed_not = self.gather_pairs([ed, edp],  **kwargs)
        out = self.subtract_layer([ed_new, ed_not],  **kwargs)
        return out



import keras as ks
from kgcnn.layers.scale import get as get_scaler
from ._model import model_disjoint
from kgcnn.layers.modules import Input
from kgcnn.models.utils import update_model_kwargs
from kgcnn.models.casting import (template_cast_output, template_cast_list_input,
                                  template_cast_list_input_docs, template_cast_output_docs)
from keras.backend import backend as backend_to_use


# Keep track of model version from commit date in literature.
__kgcnn_model_version__ = "2023-12-04"

# Supported backends
__kgcnn_model_backend_supported__ = ["tensorflow", "torch", "jax"]
if backend_to_use() not in __kgcnn_model_backend_supported__:
    raise NotImplementedError("Backend '%s' for model 'GNNFilm' is not supported." % backend_to_use())

# Implementation of GNNFilm in `keras` from paper:
# GNN-FiLM: Graph Neural Networks with Feature-wise Linear Modulation
# Marc Brockschmidt
# https://arxiv.org/abs/1906.12192


model_default = {
    "name": "GNNFilm",
    "inputs": [
        {"shape": (None,), "name": "node_attributes", "dtype": "int64"},
        {"shape": (None, ), "name": "edge_relations", "dtype": "int64"},
        {"shape": (None, 2), "name": "edge_indices", "dtype": "int64"},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_edges", "dtype": "int64"}
    ],
    "input_tensor_type": "padded",
    "input_embedding": None,  # deprecated
    "cast_disjoint_kwargs": {},
    "input_node_embedding": {"input_dim": 95, "output_dim": 64},
    "dense_relation_kwargs": {"units": 64, "num_relations": 20},
    "dense_modulation_kwargs": {"units": 64, "num_relations": 20, "activation": "sigmoid"},
    "activation_kwargs": {"activation": "swish"},
    "depth": 3,
    "verbose": 10,
    "node_pooling_kwargs": {},
    "output_embedding": 'graph',
    "output_scaling": None,
    "output_tensor_type": "padded",
    "output_to_tensor": None,  # deprecated
    "output_mlp": {"use_bias": True, "units": 1,
                   "activation": "softmax"}
}


@update_model_kwargs(model_default, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_model(inputs: list = None,
               input_tensor_type: str = None,
               cast_disjoint_kwargs: dict = None,
               input_embedding: dict = None,
               input_node_embedding: dict = None,
               depth: int = None,
               dense_relation_kwargs: dict = None,
               dense_modulation_kwargs: dict = None,
               activation_kwargs: dict = None,
               name: str = None,
               verbose: int = None,
               node_pooling_kwargs: dict = None,
               output_embedding: str = None,
               output_to_tensor: bool = None,
               output_scaling: dict = None,
               output_tensor_type: str = None,
               output_mlp: dict = None
               ):
    r"""Make `GNNFilm <https://arxiv.org/abs/1906.12192>`__ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.RGCN.model_default`.

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are  :obj:`[nodes, edge_relations, edge_indices, ...]`
    with '...' indicating mask or ID tensors following the template below.
    The edge relations do not have a feature dimension and specify the relation of each edge of type 'int'.
    Edges are actually edge single weight values which are entries of the pre-scaled adjacency matrix.

    %s

    **Model outputs**:
    The standard output template:

    %s

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layers if used.
        input_embedding (dict): Deprecated in favour of input_node_embedding etc.
        input_node_embedding (dict): Dictionary of embedding arguments unpacked in :obj:`Embedding` layers.
        depth (int): Number of graph embedding units or depth of the network.
        dense_relation_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`RelationalDense` layer.
        dense_modulation_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`RelationalDense` layer.
        activation_kwargs (dict):  Dictionary of layer arguments unpacked in :obj:`Activation` layer.
        name (str): Name of the model.
        verbose (int): Level of print output.
        node_pooling_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layer.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        output_to_tensor (bool): Whether to cast model output to :obj:`tf.Tensor`.
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".
        output_scaling (dict): Dictionary of layer arguments unpacked in scaling layers. Default is None.

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    dj_inputs = template_cast_list_input(
        model_inputs,
        input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        mask_assignment=[0, 1, 1],
        index_assignment=[None, None, 0]
    )

    n, er, disjoint_indices, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges = dj_inputs

    out = model_disjoint(
        [n, er, disjoint_indices, batch_id_node, count_nodes],
        use_node_embedding=("int" in inputs[0]['dtype']) if input_node_embedding is not None else False,
        input_node_embedding=input_node_embedding,
        depth=depth,
        dense_modulation_kwargs=dense_modulation_kwargs,
        dense_relation_kwargs=dense_relation_kwargs,
        activation_kwargs=activation_kwargs,
        output_embedding=output_embedding,
        output_mlp=output_mlp,
        node_pooling_kwargs=node_pooling_kwargs
    )

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        out = scaler(out)

    # Output embedding choice
    out = template_cast_output(
        [out, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding, output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type, cast_disjoint_kwargs=cast_disjoint_kwargs
    )

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)
    model.__kgcnn_model_version__ = __kgcnn_model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)
    return model


make_model.__doc__ = make_model.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)



from keras.layers import Add, Concatenate, Multiply, Subtract
from kgcnn.layers.gather import GatherNodes
from kgcnn.layers.aggr import AggregateLocalEdges
from kgcnn.layers.modules import Embedding
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.norm import GraphLayerNormalization
from kgcnn.layers.geom import NodePosition, EuclideanNorm, EdgeDirectionNormalized, PositionEncodingBasisLayer
from kgcnn.layers.pooling import PoolingNodes


def model_disjoint(
        inputs,
        use_node_embedding,
        use_edge_embedding,
        input_node_embedding: dict = None,
        input_edge_embedding: dict = None,
        depth: int = None,
        euclidean_norm_kwargs: dict = None,
        node_mlp_initialize: dict = None,
        use_edge_attributes: bool = None,
        edge_mlp_kwargs: dict = None,
        edge_attention_kwargs: dict = None,
        use_normalized_difference: bool = None,
        expand_distance_kwargs: dict = None,
        coord_mlp_kwargs: dict = None,
        pooling_coord_kwargs: dict = None,
        pooling_edge_kwargs: dict = None,
        node_normalize_kwargs: dict = None,
        use_node_attributes: bool = None,
        node_mlp_kwargs: dict = None,
        use_skip: bool = None,
        node_decoder_kwargs: dict = None,
        node_pooling_kwargs: dict = None,
        output_embedding: str = None,
        output_mlp: dict = None
):
    h0, x, ed, edi, batch_id_node, batch_id_edge, count_nodes, count_edges = inputs
    # Make input

    # Embedding, if no feature dimension
    if use_node_embedding:
        h0 = Embedding(**input_node_embedding)(h0)
    if use_edge_embedding:
        ed = Embedding(**input_edge_embedding)(ed)

    # Model
    h = GraphMLP(**node_mlp_initialize)([h0, batch_id_node, count_nodes]) if node_mlp_initialize else h0
    for i in range(0, depth):
        pos1, pos2 = NodePosition()([x, edi])
        diff_x = Subtract()([pos1, pos2])
        norm_x = EuclideanNorm(**euclidean_norm_kwargs)(diff_x)
        # Original code has a normalize option for coord-differences.
        if use_normalized_difference:
            diff_x = EdgeDirectionNormalized()([pos1, pos2])
        if expand_distance_kwargs:
            norm_x = PositionEncodingBasisLayer()(norm_x)

        # Edge model
        h_i, h_j = GatherNodes([0, 1], concat_axis=None)([h, edi])
        if use_edge_attributes:
            m_ij = Concatenate()([h_i, h_j, norm_x, ed])
        else:
            m_ij = Concatenate()([h_i, h_j, norm_x])
        if edge_mlp_kwargs:
            m_ij = GraphMLP(**edge_mlp_kwargs)([m_ij, batch_id_edge, count_edges])
        if edge_attention_kwargs:
            m_att = GraphMLP(**edge_attention_kwargs)([m_ij, batch_id_edge, count_edges])
            m_ij = Multiply()([m_att, m_ij])

        # Coord model
        if coord_mlp_kwargs:
            m_ij_weights = GraphMLP(**coord_mlp_kwargs)([m_ij, batch_id_edge, count_edges])
            x_trans = Multiply()([m_ij_weights, diff_x])
            agg = AggregateLocalEdges(**pooling_coord_kwargs)([h, x_trans, edi])
            x = Add()([x, agg])

        # Node model
        m_i = AggregateLocalEdges(**pooling_edge_kwargs)([h, m_ij, edi])
        if node_mlp_kwargs:
            m_i = Concatenate()([h, m_i])
            if use_node_attributes:
                m_i = Concatenate()([m_i, h0])
            m_i = GraphMLP(**node_mlp_kwargs)([m_i, batch_id_node, count_nodes])
        if node_normalize_kwargs:
            h = GraphLayerNormalization(**node_normalize_kwargs)([h, batch_id_node, count_nodes])
        if use_skip:
            h = Add()([h, m_i])
        else:
            h = m_i

    # Output embedding choice
    if node_decoder_kwargs:
        n = GraphMLP(**node_mlp_kwargs)([h, batch_id_node, count_nodes])
    else:
        n = h

    # Final step.
    if output_embedding == 'graph':
        out = PoolingNodes(**node_pooling_kwargs)([count_nodes, n, batch_id_node])
        out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `SchNet`")

    return out



import keras as ks
from keras import ops
from kgcnn.layers.gather import GatherNodes
from kgcnn.layers.aggr import AggregateLocalEdges


@ks.saving.register_keras_serializable(package='kgcnn', name='MessagePassingBase')
class MessagePassingBase(ks.layers.Layer):
    r"""Base layer for Message passing type networks. This is a general frame to implement custom message and
    update functions. The idea is to create a subclass of :obj:`MessagePassingBase` and then just implement the methods
    :obj:`message_function` and :obj:`update_nodes`. The pooling or aggregating is handled by built-in
    :obj:`AggregateLocalEdges`.

    Alternatively also :obj:`aggregate_message` could be overwritten.
    The original message passing scheme was proposed by `NMPNN <http://arxiv.org/abs/1704.01212>`__ .
    """

    def __init__(self,
                 pooling_method: str = "scatter_sum",
                 use_id_tensors: int = None,
                 **kwargs):
        r"""Initialize :obj:`MessagePassingBase` layer.

        Args:
            pooling_method (str): Aggregation method of edges. Default is "sum".
            use_id_tensors (int): Whether :obj:`call` receives graph ID information, which is passed onto message and
                aggregation function. Specifies the number of additional tensors.
        """
        super(MessagePassingBase, self).__init__(**kwargs)
        self.pooling_method = pooling_method
        self.lay_gather = GatherNodes(concat_axis=None)
        self.use_id_tensors = use_id_tensors
        self.lay_pool_default = AggregateLocalEdges(pooling_method=self.pooling_method)
    
    def build(self, input_shape):
        super(MessagePassingBase, self).build(input_shape)

    def message_function(self, inputs, **kwargs):
        r"""Defines the message function, i.e. a method the generates a message from node and edge embeddings at a
        certain depth (not considered here).

        Args:
            inputs: [nodes_in, nodes_out, edge_index]

                - nodes_in (Tensor): Receiving node embeddings of shape ([M], F)
                - nodes_out (Tensor): Sending node embeddings of shape ([M], F)
                - edges (Tensor, optional): Edge or message embeddings of shape ([M], F)

        Returns:
            Tensor: Messages for each edge of shape ([M], F)
        """
        raise NotImplementedError(
            "A method to generate messages must be implemented in sub-class of `MessagePassingBase`.")

    def aggregate_message(self, inputs, **kwargs):
        r"""Pre-defined message aggregation that uses :obj:`AggregateLocalEdges`.

        Args:
            inputs: [nodes, edges, edge_index]

                - nodes (Tensor): Node embeddings of shape ([N], F)
                - edges (Tensor): Edge or message embeddings of shape ([M], F)
                - edge_index (Tensor): Edge indices referring to nodes of shape (2, [M])

        Returns:
            Tensor: Aggregated edge embeddings per node of shape ([N], F)
        """
        return self.lay_pool_default(inputs, **kwargs)

    def update_nodes(self, inputs, **kwargs):
        r"""Defines the update function, i.e. a method that updates the node embeddings from aggregated messages.

        Args:
            inputs: [nodes, node_updates]

                - nodes (Tensor): Node embeddings (from previous step) of shape ([N], F)
                - node_updates (Tensor): Updates for nodes of shape ([N], F)

        Returns:
            Tensor: Updated node embeddings (for next step) of shape ([N], F)
        """
        raise NotImplementedError(
            "A method to update nodes must be implemented in sub-class of `MessagePassingBase`.")

    def call(self, inputs, **kwargs):
        r"""Pre-implemented standard message passing scheme using :obj:`update_nodes`, :obj:`aggregate_message` and
        :obj:`message_function`.

        Args:
            inputs: [nodes, edges, edge_index]

                - nodes (Tensor): Node embeddings of shape ([N], F)
                - edges (Tensor, optional): Edge or message embeddings of shape ([M], F)
                - edge_index (Tensor): Edge indices referring to nodes of shape (2, [M])

        Returns:
            Tensor: Updated node embeddings of shape ([N], F)
        """
        if self.use_id_tensors is not None:
            ids = inputs[-int(self.use_id_tensors):]
            inputs = inputs[:-int(self.use_id_tensors)]
        else:
            ids = []

        if len(inputs) == 2:
            nodes, edge_index = inputs[:2]
            edges = None
        else:
            nodes, edges, edge_index = inputs[:3]

        n_in, n_out = self.lay_gather([nodes, edge_index], **kwargs)

        if edges is None:
            msg = self.message_function([n_in, n_out] + ids, **kwargs)
        else:
            msg = self.message_function([n_in, n_out, edges] + ids, **kwargs)

        pool_n = self.aggregate_message([nodes, msg, edge_index], **kwargs)

        n_new = self.update_nodes([nodes, pool_n] + ids, **kwargs)
        return n_new

    def get_config(self):
        """Update config."""
        config = super(MessagePassingBase, self).get_config()
        config.update({"pooling_method": self.pooling_method, "use_id_tensors": self.use_id_tensors})
        return config


class MatMulMessages(ks.layers.Layer):
    r"""Linear transformation of edges or messages, i.e. matrix multiplication.

    The message dimension must be suitable for matrix multiplication. The actual matrix is not a trainable weight of
    this layer but passed as input.
    This was proposed by `NMPNN <http://arxiv.org/abs/1704.01212>`__ .
    For each node or edge :math:`i` the output is given by:

    .. math::

        x_i' = \mathbf{A_i} \; x_i
    """

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(MatMulMessages, self).__init__(**kwargs)

    def build(self, input_shape):
        """Build layer."""
        super(MatMulMessages, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs (list): [mat, edges]

                - mat (Tensor): Transformation matrix for each message of shape ([M], F', F).
                - edges (Tensor): Edge embeddings or messages ([M], F)

        Returns:
            Tensor: Transformation of messages by matrix multiplication of shape (batch, [M], F')
        """
        mat, e = inputs
        e_e = ops.expand_dims(e, axis=1)
        out = ops.sum(mat*e_e, axis=2)
        return out

    def get_config(self):
        """Update layer config."""
        config = super(MatMulMessages, self).get_config()
        return config
